###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import os
import json
import argparse

from cmflib import cmfquery
from cmflib.cli.command import CmdBase
from cmflib.server_interface import server_interface
from cmflib.utils.cmf_config import CmfConfig
from cmflib.utils.helper_functions import fetch_cmf_config_path
from cmflib.cmf_exception_handling import (
    TensorboardPushSuccess, 
    TensorboardPushFailure, 
    MlmdFilePushSuccess,
    ExecutionsAlreadyExists,
    FileNotFound,
    ExecutionUUIDNotFound,
    PipelineNotFound,
    UpdateCmfVersion,
    CmfServerNotAvailable,
    InternalServerError,
    InvalidTensorboardFilePath,
    MissingArgument,
    DuplicateArgumentNotAllowed
)

# This class pushes mlmd file to cmf-server
class CmdMetadataPush(CmdBase):

    # Create a function to search for files into multiple directories
    def search_files(self, file_list, *directories):
        found_files = {}
        for directory in directories:
            abs_dir = os.path.abspath(directory)
            for file_name in file_list:
                if isinstance(file_name, str):
                    file_path = os.path.join(abs_dir, file_name)
                    if os.path.isfile(file_path):
                        found_files[file_name] = file_path
        return found_files

    def run(self, live):
        current_directory = mlmd_directory = os.getcwd()
        mlmd_file_name = "./mlmd"
        
        output, cmf_config_path = fetch_cmf_config_path()
        attr_dict = CmfConfig.read_config(cmf_config_path)
        url = attr_dict.get("cmf-server-url", "http://127.0.0.1:80")
        #print(attr_dict)

        cmd_args = {
            "file_name": self.args.file_name,
            "pipeline_name": self.args.pipeline_name,
            "execution_uuid": self.args.execution_uuid,
            "tensorboard_path": self.args.tensorboard_path
        }  
        for arg_name, arg_value in cmd_args.items():
            if arg_value:
                if arg_value[0] == "":
                    raise MissingArgument(arg_name)
                elif len(arg_value) > 1:
                    raise DuplicateArgumentNotAllowed(arg_name,("-"+arg_name[0]))
         

        current_directory = os.getcwd()
        if not self.args.file_name:         # If self.args.file_name is None or an empty list ([]). 
            mlmd_file_name = "./mlmd"       # Default path for mlmd file name.
        else:
            mlmd_file_name = self.args.file_name[0].strip()
            if mlmd_file_name == "mlmd":
                mlmd_file_name = "./mlmd"
        
        current_directory = os.path.dirname(mlmd_file_name)
        
        # checks if mlmd file is present in current directory or given directory
        if not os.path.exists(mlmd_file_name):
            raise FileNotFound(mlmd_file_name, mlmd_directory)

        query = cmfquery.CmfQuery(mlmd_file_name)
        status_code = 0

        # Checks if pipeline name exists
        pipeline_name = self.args.pipeline_name[0]
        if pipeline_name in query.get_pipeline_names():
            print("metadata push started")
            print("........................................")
            # converts mlmd file to json format
            json_payload = query.dumptojson(pipeline_name, None)

            # checks if execution is given by user
            if not self.args.execution_uuid:         # If self.args.execution_uuid is None or an empty list ([]).
                exec_uuid = None
                response = server_interface.call_mlmd_push(json_payload, url, exec_uuid, pipeline_name)
            else:
                execution_flag = 0
                exec_uuid = self.args.execution_uuid[0]
                mlmd_data = json.loads(json_payload)["Pipeline"]
                # checks if given execution present in mlmd
                for i in mlmd_data[0]["stages"]:
                    for j in i["executions"]:
                        # created exec_uuid of list if multiple uuid present for single execution.
                        # for eg: f9da581c-d16c-11ef-9809-9350156ed1ac,32f17f4a-d16d-11ef-9809-9350156ed1ac
                        uuid_list = j['properties']['Execution_uuid'].split(",")
                        # check if user specified exec_uuid exists inside local mlmd
                        if exec_uuid in uuid_list: 
                            execution_flag = 1
                            # calling mlmd_push api to push mlmd_data = json.loads(json_payload)["Pipeline"]
                # checks if given execution present in mlmdmlmd file to cmf-server
                            response = server_interface.call_mlmd_push(
                                json_payload, url, exec_uuid, pipeline_name
                            )
                            break
                if execution_flag == 0:
                    raise ExecutionUUIDNotFound(exec_uuid)
            status_code = response.status_code

            # we need to push the python env files only after the mlmd push has succeded 
            # otherwise there is no use of those python env files on cmf-server

            if status_code==422 and response.json()["status"]=="version_update":
                raise UpdateCmfVersion
            elif status_code == 400:
                raise CmfServerNotAvailable
            elif status_code == 500:
                raise InternalServerError
            elif status_code == 200:
                # the only question remains how we want to percieve the failure of upload of the python env files 
                # for now, it is considered as non-consequential.
                # that means it's failure/success doesn't matter.
                # however, we will be keeping the record of the status code.
                
                # Getting all executions df to get the custom property 'Python_Env'
                executions = query.get_all_executions_in_pipeline(pipeline_name)
                if not executions.empty:
                    if 'custom_properties_Python_Env' in executions.columns:
                        list_of_env_files = executions['custom_properties_Python_Env'].drop_duplicates().tolist()
                        # This is a validation step to suppress errors in case user is pushing the mlmd
                        # from a directory in which 'cmf_artifacts/Python_Env_hash.txt' is not present.
                        # Find the valid file paths.  
                        found_files = self.search_files(list_of_env_files, current_directory, mlmd_directory) 

                        # push valid files on cmf-server
                        if found_files:
                            for name, path in found_files.items():
                                env_response = server_interface.call_python_env(url, name, path)
                                # keeping record of status but this won't affect the mlmd success.
                                print(env_response.json())

                
                # get labels for every artifact
                artifacts = query.get_all_artifacts_by_context(pipeline_name)
                if not artifacts.empty:
                    if "custom_properties_labels_uri" in artifacts.columns:
                        labels_with_uri = artifacts["custom_properties_labels_uri"].dropna().drop_duplicates().tolist()
                        # every artifacts can contain multiple labels. 
                        # when 'labels' column has more than one label, it looks as follows 
                        # labels = "labels.csv, labels1.csv, labels2.csv"
                        for l in labels_with_uri:
                            labels = l.split(",")
                            for label in labels:
                                label_name = label.split(":")[1]
                                path = os.getcwd() +"/"+ label.split(":")[0]
                                label_response = server_interface.call_label(url, label_name, path)
                                print(label_response.json())

                output = ""
                display_output = ""
                if response.json()['status']=="success":
                    display_output = "mlmd is successfully pushed."
                    output = MlmdFilePushSuccess(mlmd_file_name)
                if response.json()["status"]=="exists":
                    display_output = "Executions already exists."
                    output = ExecutionsAlreadyExists()
                if not self.args.tensorboard_path:
                    return output
                print(display_output)
                # /tensorboard api call is done only if mlmd push is successfully completed
                # tensorboard parameter is passed
                print("......................................")
                print("tensorboard logs upload started!!")
                print("......................................")


                tensorboard = self.args.tensorboard_path[0]
                # check if the path provided is for a file
                if os.path.isfile(tensorboard):
                    file_name = os.path.basename(tensorboard)
                    tresponse = server_interface.call_tensorboard(url, pipeline_name, file_name, tensorboard)
                    tstatus_code = tresponse.status_code
                    if tstatus_code == 200:
                        # give status code as success
                        return TensorboardPushSuccess(file_name)
                    else:
                        # give status code as failure 
                        return TensorboardPushFailure(file_name,tresponse.text)
                # If path provided is a directory
                elif os.path.isdir(tensorboard):
                    # Recursively push all files and subdirectories
                    for root, dirs, files in os.walk(tensorboard):
                        for file_name in files:
                            file_path = os.path.join(root, file_name)
                            relative_path = os.path.relpath(file_path, tensorboard)
                            tresponse = server_interface.call_tensorboard(url, pipeline_name, relative_path, file_path)
                            if tresponse.status_code == 200:
                                print(f"tensorboard logs: File {file_name} uploaded successfully.")
                            else:
                                # give status as failure
                                return TensorboardPushFailure(file_name,tresponse.text)
                    return TensorboardPushSuccess()
                else:
                    return InvalidTensorboardFilePath()   
            else:
                return "ERROR: Status Code = {status_code}. Unable to push mlmd."
        else:
            raise PipelineNotFound(pipeline_name)


def add_parser(subparsers, parent_parser):
    PUSH_HELP = "Push user-generated metadata file to server to create one single metadata file for all the pipelines."

    parser = subparsers.add_parser(
        "push",
        parents=[parent_parser],
        description="Push user's metadata file to cmf-server.",
        help=PUSH_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "-p",
        "--pipeline_name",
        required=True,
        action="append",
        help="Specify Pipeline name.",
        metavar="<pipeline_name>",
    )

    parser.add_argument(
        "-f", 
        "--file_name", 
        action="append",
        help="Specify input metadata file name.", 
        metavar="<file_name>"
    )

    parser.add_argument(
        "-e",
        "--execution_uuid",
        action="append",
        help="Specify Execution uuid.",
        metavar="<exec_uuid>",
    )

    parser.add_argument(
        "-t",
        "--tensorboard_path",
        action="append",
        help="Specify path to tensorboard logs for the pipeline.",
        metavar="<tensorboard_path>"
    )

    parser.set_defaults(func=CmdMetadataPush)



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
# Error: Skipping analyzing "dvc.api": module is installed, but missing library stubs or py.typed marker
from dvc.api import DVCFileSystem    # type: ignore

class LocalArtifacts():
    """
        Initialize the LocalArtifacts class with local repo url.
        This class downloads one local artifact at a time and if the passed artifact is a directory 
        then, it downloads all the files from the directory 

        Args:
            dvc_config_op (dict): Dictionary containing local url (remote.local.url).
        """
    
    def __init__(self, dvc_config_op):
        self.fs = DVCFileSystem(
                dvc_config_op["remote.local-storage.url"]
            )  # dvc_config_op[1] is file system path - "/path/to/local/repository"
        
    def download_file(
        self,
        current_directory: str,
        object_name: str,
        download_loc: str,
    ):
        """
        Download a single file from an local repository.

        Args:
            current_directory (str): The current working directory.
            object_name (str): Key (path) of the file in the local repo.
            download_loc (str): Local path where the file should be downloaded.

        Returns:
            tuple: (object_name, download_loc, status) where status indicates success (True) or failure (False).
        """
        # get_file() only creates file, to put artifacts in proper directory, subfolders are required.
        # download_loc = contains absolute path of the file with file name and extension

        # Create necessary directories for the download location.
        dir_path = ""
        if "/" in download_loc:
            dir_path, _ = download_loc.rsplit("/", 1)
        if dir_path != "":
            os.makedirs(dir_path, mode=0o777, exist_ok=True)  # creating subfolders if needed
        
        try:
            # get_file() returns none when file gets downloaded.
            response = self.fs.get_file(object_name, download_loc)

            # Check if the response indicates success.
            if response == None:  
                return object_name, download_loc, True
            return  object_name, download_loc, False
        except Exception as e:
            return  object_name, download_loc, False


    def download_directory(
        self,
        current_directory: str,
        object_name: str,
        download_loc: str,
    ):
        """
        Download a directory from an local repo using its .dir metadata object.

        Args:
            current_directory (str): The current working directory .
            object_name (str): Key (path) of the .dir object in the local repository.
            download_loc (str): Local directory path where the directory should be downloaded.

        Returns:
            tuple: (total_files_in_directory, files_downloaded, status) where status indicates success (True) or failure (False).
        """

        # get_file() only creates file, to put artifacts in proper directory, subfolders are required.
        # download_loc = contains absolute path of the file with file name and extension
        dir_path = ""
        if "/" in download_loc:
            dir_path, _ = download_loc.rsplit("/", 1)
        if dir_path != "":
            os.makedirs(dir_path, mode=0o777, exist_ok=True)  # creating subfolders if needed

        """"
        if object_name ends with .dir - it is a directory.
        we download .dir object with 'temp_dir' and remove 
        this after all the files from this .dir object is downloaded.
        """
        # in case of .dir, download_loc is a absolute path for a folder
        os.makedirs(download_loc, mode=0o777, exist_ok=True)
        total_files_in_directory = 0
        files_downloaded = 0

        # Temporary file to download the .dir metadata object.
        temp_dir = f"{download_loc}/dir"
        try:
            # Download the .dir file containing metadata about tracked files.
            response = self.fs.get_file(object_name, temp_dir)
            
            with open(temp_dir, 'r') as file:
                tracked_files = eval(file.read())

            # removing temp_dir
            if os.path.exists(temp_dir):
                os.remove(temp_dir)

            """
            object_name = "files/md5/9b/9a458ac0b534f088a47c2b68bae479.dir" 
            contains the path of the .dir on the artifact repo
            we need to remove the hash of the .dir from the object_name
            which will leave us with the artifact repo path
            """
            repo_path = "/".join(object_name.split("/")[:-2])

            obj = True
            for file_info in tracked_files:
                total_files_in_directory += 1
                relpath = file_info['relpath']
                md5_val = file_info['md5']
                # md5_val = a237457aa730c396e5acdbc5a64c8453
                # we need a2/37457aa730c396e5acdbc5a64c8453
                formatted_md5 = md5_val[:2] + '/' + md5_val[2:]
                temp_object_name = f"{repo_path}/{formatted_md5}"
                temp_download_loc = f"{download_loc}/{relpath}"
                try:
                    obj = self.fs.get_file(temp_object_name, temp_download_loc)
                    if obj == None: 
                        files_downloaded += 1
                        print(f"object {temp_object_name} downloaded at {temp_download_loc}.")
                    else:
                        print(f"object {temp_object_name} is not downloaded.")
                # this exception is for get_file() function for temp_object_name
                except Exception as e:
                    print(f"object {temp_object_name} is not downloaded.")

            # total_files - files_downloaded gives us the number of files which are failed to download
            if (total_files_in_directory - files_downloaded) == 0:   
                return total_files_in_directory, files_downloaded, True
            return total_files_in_directory, files_downloaded, False  
        # this exception is for get_file() function for object_name
        except Exception as e:
            print(f"object {object_name} is not downloaded.")
            # We usually don't count .dir as a file while counting total_files_in_directory.
            # However, here we failed to download the .dir folder itself. 
            # So we need to make, total_files_in_directory = 1
            total_files_in_directory = 1 
            return total_files_in_directory, files_downloaded, False

        # sometimes we get TypeError as an execption, however investiagtion for the exact scenarios is pending



from ray import tune
from ray.tune import Callback
from cmflib import cmf
import heapq

class CmfRayLogger(Callback):
    #id_count = 1
    
    def __init__(self, pipeline_name, file_path, pipeline_stage, data_dir = None, metric = 'accuracy', order = 'max', top_n=5 ):
        """
        pipeline_name: The name of the CMF Pipeline
        file_path: The path to metadata file
        pipeline_stage: The name for the stage of cmf_pipeline
        data_dir = Directory/File for data that is logged
        metric: The metric to track (e.g., 'accuracy', 'loss')
        order: 'max' for maximum, 'min' for minimum
        top_n: Number of top results to keep
        """
        self.pipeline_name = pipeline_name
        self.file_path = file_path
        self.pipeline_stage = pipeline_stage
        self.cmf_obj = {}
        self.cmf_run = {}
        self.data_dir = data_dir
        self.metric = metric
        self.order = order
        self.top_n = top_n
        
        # Initialize heap based on user-defined order
        self.heap = []
        self.heap_comparator = -1 if self.order == 'max' else 1
        
        # Dictionary to track best metric and model for each trial
        self.best_metric_values = {}
        self.best_models = {}
        self.execution_ids = {}

    def on_trial_start(self, iteration, trials, trial, **info):
        trial_id = trial.trial_id
        trial_config = trial.config
        print(f"CMF Logging Started for Trial {trial_id}")
        self.cmf_obj[trial_id] = cmf.Cmf(filepath = self.file_path, pipeline_name = self.pipeline_name)
        _ = self.cmf_obj[trial_id].create_context(pipeline_stage = self.pipeline_stage)
        execution_id = self.cmf_obj[trial_id].create_execution(execution_type=f"Trial_{trial_id}",
                                            create_new_execution = False,
                                            custom_properties = {'Configuration': trial_config})
        
        # Store the execution_type which will be used to update the execution later
        self.execution_ids[trial_id] = execution_id.id

        if self.data_dir:
            _ = self.cmf_obj[trial_id].log_dataset(url = str(self.data_dir), event = 'input')
        
        self.best_metric_values[trial_id] = None
        self.best_models[trial_id] = None

    def on_trial_result(self, iteration, trials, trial, result, **info):
        trial_id = trial.trial_id
        trial_config = trial.config
        trial_result = trial.last_result
        curr_res = result
        #print(f'In Trial Results')
        print(f'Trial_ID: {trial_id}')
        print(f'curr_results is {curr_res}')
        print(f'Logging results to CMF with metric name Trial_{trial_id}_metrics')
        #if trial_id in self.execution_id:
        #    _ = self.metawriter.update_execution(int(self.execution_id[trial_id]))
        _ = self.cmf_obj[trial_id].log_metric(metrics_name = f"Trial_{trial_id}_metrics",
                                      custom_properties = {'Output': curr_res})
        self.cmf_run[trial_id] = True
        
        # Track the current metric value and model path (if available)
        metric_value = curr_res.get(self.metric, None)
        model_path = curr_res.get('model_path', None) # Track the model path if available
        
        
        # Update best metric and model for the trial if necessary
        if metric_value is not None:
            if self.best_metric_values[trial_id] is None:
                self.best_metric_values[trial_id] = metric_value
                self.best_models[trial_id] = model_path
            else:
                # Update best metric based on order (max/min)
                if ((self.order == 'max' and metric_value > self.best_metric_values[trial_id]) or 
                    (self.order == 'min' and metric_value < self.best_metric_values[trial_id])):
                    self.best_metric_values[trial_id] = metric_value
                    self.best_models[trial_id] = model_path
        
    def on_trial_complete(self, iteration, trials, trial, **info):
        trial_id = trial.trial_id
        trial_config = trial.config
        trial_result = trial.last_result

        best_metric_value = self.best_metric_values.get(trial_id, None)
        best_model_path = self.best_models.get(trial_id, None)
        execution_id = self.execution_ids[trial_id]
        
        if best_metric_value is not None:
            # Push the best value of the trial and its corresponding model into the heap
            heapq.heappush(self.heap, (self.heap_comparator * best_metric_value, trial_id, best_metric_value, best_model_path,
                                       execution_id))
            if len(self.heap) > self.top_n:
                heapq.heappop(self.heap)  # Maintain top_n elements in the heap
        
            
            _ = self.cmf_obj[trial_id].log_execution_metrics(metrics_name=f"Best_{self.metric}_Trial_{trial_id}",
                                                             custom_properties={f'Best_{self.metric}': best_metric_value,
                                                                                'execution_id': execution_id}
                                                            )

        # Commit the metrics for the trial and log its final state
        print(f"Trial {trial_id} completed, Commiting to CMF: with name Trial_{trial_id}_metrics")
        print()

        _ = self.cmf_obj[trial_id].commit_metrics(f"Trial_{trial_id}_metrics")
        
        if best_model_path:
            _ = self.cmf_obj[trial_id].log_model(path=best_model_path, 
                                                 event='input', 
                                                 model_name=f"{trial_id}_model")
        
    def on_trial_error(self, iteration, trials, trial, **info):
        trial_id = trial.trial_id
        trial_config = trial.config
        trial_result = trial.last_result

        print(f"An error occured with Trial {trial_id}, Not commiting anything to cmf")
        if self.cmf_run[trial_id]:
            _ = self.cmf_obj[trial_id].commit_metrics(f"Trial_{trial_id}_metrics")
            _ = self.cmf_obj[trial_id].log_execution_metrics(metrics_name = f"Trial_{trial_id}_Result",
                                      custom_properties = {'Result': '-inf'})
            
    def on_experiment_end(self, trials, **info):
        """
        This function is called at the end of the experiment to log the top 'n' results and update execution
        with {'in_top_n': True} for each of the top trials.
        """
        print(f"Marking top {self.top_n} trials with 'in_top_n = True' at experiment end.")

        # Log the top 'n' trials from the heap
        while self.heap:
            _, top_trial_id, top_metric_value, top_model_path, top_execution_id = heapq.heappop(self.heap)
            print(f"Top trial: {top_trial_id} with {self.metric}: {top_metric_value}")

            # Update the execution for this trial with 'in_top_n = True'
            _ = self.cmf_obj[top_trial_id].update_execution(
                execution_id=top_execution_id,
                custom_properties={'in_top_n': True}
            )

            print(f"Execution {top_execution_id} for Trial {top_trial_id} updated with 'in_top_n = True'.")


def modify_arti_name(arti_name, type):
    # artifact_name optimization based on artifact type.["Dataset","Model","Metrics"]
    try:
        name = ""

        if type == "Metrics" or type == "Model" or type == "Dataset":
            # Metrics   metrics:7bea36fc-8b99-11ef-abea-ddaa7ef0aa99:13  -----------> ['metrics', '7bea36fc-8b99-11ef-abea-ddaa7ef0aa99', '13']
            # Dataset   artifacts/data.xml.gz:236d9502e0283d91f689d7038b8508a2  -----------> ['artifacts/data.xml.gz', '236d9502e0283d91f689d7038b8508a2']
            # Model   artifacts/model/model.pkl:4c48f23acd14d20ebba0352f4b5f55e8:9  ------> ['artifacts/model/model.pkl', '4c48f23acd14d20ebba0352f4b5f55e8', '9']
            split_by_colon = arti_name.split(':')

        if type == "Dataslice" or type == "Step_Metrics":
            # Step_Metrics   cmf_artifacts/5a2f6686-8b99-11ef-abea-ddaa7ef0aa99/metrics/training_metrics:46fd4d02f72dee5fc88b0cf9aa908ed5:15:744ad0be-8b99-11ef-abea-ddaa7ef0aa99 
            # ---> 5a2f6686-8b99-11ef-abea-ddaa7ef0aa99/metrics/training_metrics:46fd4d02f72dee5fc88b0cf9aa908ed5:15:744ad0be-8b99-11ef-abea-ddaa7ef0aa99
            # Dataslice   cmf_artifacts/c1e542fc-8ba1-11ef-abea-ddaa7ef0aa99/dataslice/slice-1:059136b3b35fc4b58cf13f73e4564b9b
            # ----> "c1e542fc-8ba1-11ef-abea-ddaa7ef0aa99/dataslice/slice-1:059136b3b35fc4b58cf13f73e4564b9b"
            split_by_slash = arti_name.split('/', 1)[1] #remove cmf_artifacts/

        if type ==  "Dataset" or type == "Step_Metrics":
            # Dataset   artifacts/data.xml.gz:236d9502e0283d91f689d7038b8508a2  -----------> ["artifacts/data.xml.gz","236d9502e0283d91f689d7038b8508a2"]
            # Step_Metrics   cmf_artifacts/5a2f6686-8b99-11ef-abea-ddaa7ef0aa99/metrics/training_metrics:46fd4d02f72dee5fc88b0cf9aa908ed5:15:744ad0be-8b99-11ef-abea-ddaa7ef0aa99 
            # ----> ["cmf_artifacts/5a2f6686-8b99-11ef-abea-ddaa7ef0aa99/metrics/training_metrics", "46fd4d02f72dee5fc88b0cf9aa908ed5", "15" "744ad0be-8b99-11ef-abea-ddaa7ef0aa99"]
            rsplit_by_colon = arti_name.rsplit(':')
       
        if type == "Metrics" :   
            # split_by_colon = ["metrics","7bea36fc-8b99-11ef-abea-ddaa7ef0aa99","13"] ----> "metrics:7bea:13"
            # name = "metrics:7bea:13"
            name = f"{split_by_colon[0]}:{split_by_colon[1][:4]}:{split_by_colon[2]}"

        elif type == "Model":
            # split_by_colon = ["artifacts/model/model.pkl", "4c48f23acd14d20ebba0352f4b5f55e8", "9"]
            # split_by_colon[-3].split("/")[-1] --> "model.pkl"
            # split_by_colon[-2][:4] --> "4c48"
            # name = "model.pkl:4c48"
            name = split_by_colon[-3].split("/")[-1] + ":" + split_by_colon[-2][:4]

        elif type == "Dataset":
            # Example artifacts/data.xml.gz:236d9502e0283d91f689d7038b8508a2 -> "data.xml.gz:236d"
            # rsplit_by_colon --> ["artifacts/data.xml.gz","236d9502e0283d91f689d7038b8508a2"] 
            # rsplit_by_colon[0].split("/")[-1] ---> artifacts/data.xml.gz ---> "data.xml.gz"
            # split_by_colon[-1][:4] ---> ["artifacts/data.xml.gz","236d9502e0283d91f689d7038b8508a2"]  ---> "236d"
            # name = "data.xml.gz:236d"
            name = rsplit_by_colon[0].split("/")[-1] + ":" +  split_by_colon[-1][:4]

        elif type == "Dataslice":
            # split_by_slash = "c1e542fc-8ba1-11ef-abea-ddaa7ef0aa99/dataslice/slice-1:059136b3b35fc4b58cf13f73e4564b9b"
            # data = ["c1e542fc-8ba1-11ef-abea-ddaa7ef0aa99/dataslice/slice-1", "059136b3b35fc4b58cf13f73e4564b9b"]
            # name = "c1e542fc-8ba1-11ef-abea-ddaa7ef0aa99/dataslice/slice-1:0591"
            data = split_by_slash.rsplit(":",-1)
            name = data[0] + ":" + data[-1][:4]

        elif type == "Step_Metrics":
            # split_by_slash = 5a2f6686-8b99-11ef-abea-ddaa7ef0aa99/metrics/training_metrics:46fd4d02f72dee5fc88b0cf9aa908ed5:15:744ad0be-8b99-11ef-abea-ddaa7ef0aa99 
            # split_by_slash.rsplit(":",-3)[0] = "5a2f6686-8b99-11ef-abea-ddaa7ef0aa99/metrics/training_metrics"
            # rsplit_by_colon = ["cmf_artifacts/5a2f6686-8b99-11ef-abea-ddaa7ef0aa99/metrics/training_metrics", "46fd4d02f72dee5fc88b0cf9aa908ed5", "15", "744ad0be-8b99-11ef-abea-ddaa7ef0aa99"]
            # rsplit_by_colon[-3][:4] = "46fd"
            # rsplit_by_colon[-2] = "15"
            # rsplit_by_colon[-1][:4] = "744a"
            # name = "5a2f6686-8b99-11ef-abea-ddaa7ef0aa99/metrics/training_metrics:46fd:15:744a"
            name = split_by_slash.rsplit(":",-3)[0] + ":" + rsplit_by_colon[-3][:4] + ":" + rsplit_by_colon[-2] + ":" + rsplit_by_colon[-1][:4]
        else:
            name = arti_name  
    except Exception as e:
        print(f"Error parsing artifact name: {e}")
        name = arti_name  # Fallback to the original arti_name in case of error
    return name
 


###
# Copyright (2024) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import argparse
import os
import re

from cmflib import cmfquery
from cmflib.cli.utils import check_minio_server, find_root
from cmflib.utils.helper_functions import generate_osdf_token, branch_exists
from cmflib.utils.dvc_config import DvcConfig
from cmflib.dvc_wrapper import dvc_add_attribute
from cmflib.utils.cmf_config import CmfConfig
from cmflib.cli.command import CmdBase
from cmflib.dvc_wrapper import git_get_repo, git_get_pull, git_get_push, git_get_branch, dvc_push
from cmflib.commands.artifact.push import CmdArtifactPush
from cmflib.commands.metadata.push import CmdMetadataPush
from cmflib.cmf_exception_handling import (
    MsgSuccess, 
    MsgFailure, 
    ArtifactPushSuccess, 
    Minios3ServerInactive, 
    CmfNotConfigured, 
    FileNotFound,
    ExecutionUUIDNotFound,
    MissingArgument,
    DuplicateArgumentNotAllowed,
)


class CmdRepoPush(CmdBase):   
    def git_push(self):
        # Getting GitHub URL from cmf init command
        url = git_get_repo()
        # Example url = https://github.com/ABC/my-repo
        # Check if the URL is a GitHub URL
        if "github.com" not in url:
            raise MsgFailure(msg_str="The repository URL is not a GitHub URL.")
        
        # Extracting the repository owner and name from the URL
        # repo_owner = ABC, repo_name = my-repo
        url_parts = url.split("/")
        repo_owner = url_parts[-2]
        repo_name = url_parts[-1]
        
        # Getting the current branch name
        branch_name = git_get_branch()[0]
        
        # Check whether the branch exists in the GitHub repository
        if branch_exists(repo_owner, repo_name, branch_name):
            # Pull the latest changes from the branch
            stdout, stderr, returncode = git_get_pull(branch_name)
            if returncode != 0:
                raise MsgFailure(msg_str=f"Git pull failed: {stderr}")
            print(stdout)
        
        # Push the changes to the branch
        stdout, stderr, returncode = git_get_push(branch_name)
        if returncode != 0:
            raise MsgFailure(msg_str=f"Git push failed: {stderr}")
        
        return MsgSuccess(msg_str="cmf repo push command executed successfully.")
    
    def artifact_push(self, live):
        """
        Pushes artifacts to the remote storage.

        Raises:
            MissingArgument: If a required argument is missing.
            DuplicateArgumentNotAllowed: If a duplicate argument is provided.
            CmfNotConfigured: If the .cmfconfig file is not configured.
            Minios3ServerInactive: If the Minio server is inactive.
            FileNotFound: If the mlmd file does not exist.
            ExecutionUUIDNotFound: If the execution UUID is not found in the pipeline.

        Returns:
            ArtifactPushSuccess: If the artifacts are successfully pushed to the remote storage.
        """
        cmd_args = {
            "file_name": self.args.file_name,
            "pipeline_name": self.args.pipeline_name,
            "execution_uuid": self.args.execution_uuid,
            "tensorboad": self.args.tensorboard,
            "jobs": self.args.jobs
        }
        # Validates the command arguments.
        for arg_name, arg_value in cmd_args.items():
            if arg_value:
                if arg_value[0] == "":
                    raise MissingArgument(arg_name)
                # Check if the argument is a list and has more than one value
                elif len(arg_value) > 1:
                    raise DuplicateArgumentNotAllowed(arg_name,("-"+arg_name[0]))
                
        result = ""

        # Checks the DVC configuration and the presence of the .cmfconfig file.
        dvc_config_op = DvcConfig.get_dvc_config()
        cmf_config_file = os.environ.get("CONFIG_FILE", ".cmfconfig")

        # find root_dir of .cmfconfig
        output = find_root(cmf_config_file)

        # in case, there is no .cmfconfig file
        if output.find("'cmf' is not configured.") != -1:
            raise CmfNotConfigured(output)

        # Verifies the Minio server status if the remote is Minio.
        out_msg = check_minio_server(dvc_config_op)
        if dvc_config_op["core.remote"] == "minio" and out_msg != "SUCCESS":
            raise Minios3ServerInactive()
        
        # If user has not specified the number of jobs or jobs is not a digit, set it to 4 * cpu_count()
        num_jobs = int(self.args.jobs[0]) if self.args.jobs and self.args.jobs[0].isdigit() else 4 * os.cpu_count()
        
        # If the remote is OSDF, generate a dynamic password and update the DVC configuration.
        if dvc_config_op["core.remote"] == "osdf":
            config_file_path = os.path.join(output, cmf_config_file)
            cmf_config={}
            cmf_config=CmfConfig.read_config(config_file_path)
            #print("key_id="+cmf_config["osdf-key_id"])
            dynamic_password = generate_osdf_token(cmf_config["osdf-key_id"],cmf_config["osdf-key_path"],cmf_config["osdf-key_issuer"])
            #print("Dynamic Password"+dynamic_password)
            dvc_add_attribute(dvc_config_op["core.remote"],"password",dynamic_password)
            #The Push URL will be something like: https://<Path>/files/md5/[First Two of MD5 Hash]
            result = dvc_push(num_jobs)
            return result

        # Determines the mlmd file name and checks its existence.
        current_directory = os.getcwd()
        if not self.args.file_name:         # If self.args.file_name is None or an empty list ([]). 
            mlmd_file_name = "./mlmd"       # Default path for mlmd file name.
        else:
            mlmd_file_name = self.args.file_name[0]
            if mlmd_file_name == "mlmd":
                mlmd_file_name = "./mlmd"
            current_directory = os.path.dirname(mlmd_file_name)
        if not os.path.exists(mlmd_file_name):   #checking if MLMD files exists
            raise FileNotFound(mlmd_file_name, current_directory)
        
        # Creates a CmfQuery object and retrieves all executions in the specified pipeline.
        query = cmfquery.CmfQuery(mlmd_file_name)
        names = []
        isExecUuid = False
        df = query.get_all_executions_in_pipeline(self.args.pipeline_name[0])

        # checking if execution_uuid exists in the df
        for index, row in df.iterrows():
            if row['Execution_uuid'] == self.args.execution_uuid[0] or self.args.execution_uuid[0] in row['Execution_uuid']:
                isExecUuid = True
                break
        
        if not isExecUuid:
            raise ExecutionUUIDNotFound(self.args.execution_uuid[0])
        
        # fetching execution id from df based on execution_uuid 
        exec_id_df = df[df['Execution_uuid'].apply(lambda x: self.args.execution_uuid[0] in x.split(","))]['id'] 
        exec_id = int(exec_id_df.iloc[0])
        
        artifacts = query.get_all_artifacts_for_execution(exec_id)  # getting all artifacts based on execution id
        # dropping artifact with type 'metrics' as metrics doesn't have physical file
        if not artifacts.empty:
            artifacts = artifacts[artifacts['type'] != 'Metrics']
            # adding .dvc at the end of every file as it is needed for pull
            artifacts['name'] = artifacts['name'].apply(lambda name: f"{name.split(':')[0]}.dvc")
            names.extend(artifacts['name'].tolist())

        final_list = []
        for file in set(names):
            # checking if the .dvc exists
            if os.path.exists(file):
                final_list.append(file)
            # checking if the .dvc exists in user's project working directory
            elif os.path.isabs(file):
                    file = re.split("/",file)[-1]
                    file = os.path.join(os.getcwd(), file)
                    if os.path.exists(file):
                        final_list.append(file)
            else:
                # not adding the .dvc to the final list in case .dvc doesn't exists in both the places
                pass
        result = dvc_push(num_jobs, list(final_list))
        return ArtifactPushSuccess(result)
        

    def run(self, live):
        print("Executing cmf artifact push command..")
        if(self.args.execution_uuid):
            # If an execution uuid exists, push the artifacts associated with that execution. 
            artifact_push_result = self.artifact_push(live)
        else:
            # Pushing all artifacts. 
            artifact_push_instance = CmdArtifactPush(self.args)
            artifact_push_result = artifact_push_instance.run(live)

        if artifact_push_result.status == "success":
            print("Executing cmf metadata push command..")
            metadata_push_instance = CmdMetadataPush(self.args)
            metadata_push_result = metadata_push_instance.run(live)
            if metadata_push_result.status == "success":
                print(metadata_push_result.handle())  # Print the message returned by the handle() method of the metadata_push_result object.
                print("Executing git push command..")
                live.stop()
                return self.git_push()
    

def add_parser(subparsers, parent_parser):
    PUSH_HELP = "Push artifacts, metadata files, and source code to the user's artifact repository, cmf-server, and git respectively."

    parser = subparsers.add_parser(
        "push",
        parents=[parent_parser],
        description="Push artifacts, metadata files, and source code to the user's artifact repository, cmf-server, and git respectively.",
        help=PUSH_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "-p",
        "--pipeline_name",
        required=True,
        action="append",
        help="Specify Pipeline name.",
        metavar="<pipeline_name>",
    )

    parser.add_argument(
        "-f", 
        "--file_name", 
        action="append",
        help="Specify input metadata file name.", 
        metavar="<file_name>"
    )

    parser.add_argument(
        "-e",
        "--execution_uuid",
        action="append",
        help="Specify Execution uuid.",
        default=None,
        metavar="<exec_uuid>",
    )

    parser.add_argument(
        "-t",
        "--tensorboard_path",
        action="append",
        help="Specify path to tensorboard logs for the pipeline.",
        metavar="<tensorboard_path>"
    )

    parser.add_argument(
        "-j",
        "--jobs",
        action="append",
        help="Number of parallel jobs for uploading artifacts to remote storage. Default is 4 * cpu_count(). Increasing jobs may speed up uploads but will use more resources.",
        metavar="<jobs>"
    )

    parser.set_defaults(func=CmdRepoPush)



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import argparse

from cmflib.cli.command import CmdBase
from cmflib.utils.cmf_config import CmfConfig
from cmflib.cmf_exception_handling import CmfInitShow
from cmflib.utils.helper_functions import fetch_cmf_config_path

class CmdInitShow(CmdBase):
    def run(self, live):
        output, config_file_path = fetch_cmf_config_path()
        attr_dict = CmfConfig.read_config(config_file_path)
        # Combine the two dictionaries
        combined_dict = output | attr_dict
        attr_list = []
        for key, value in combined_dict.items():
            temp_str = f"{key} = {value}"
            attr_list.append(temp_str)
        attr_str = "\n".join(attr_list)
        return CmfInitShow(attr_str)


def add_parser(subparsers, parent_parser):
    HELP = "Show current CMF configuration."

    parser = subparsers.add_parser(
        "show",
        parents=[parent_parser],
        description="This command shows current cmf configuration including cmf server ip.",
        help=HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.set_defaults(func=CmdInitShow)



from abc import ABC, abstractmethod
from ml_metadata.metadata_store import metadata_store

class CmfStore(ABC):

    def __init__(self, config):
        self.config = config
        super().__init__()

    @abstractmethod
    def connect(self) -> metadata_store.MetadataStore:
        cmf_store = metadata_store.MetadataStore(self.config)
        return cmf_store



from sqlalchemy.ext.asyncio import AsyncSession
from fastapi import Depends
from server.app.db.dbconfig import get_db
from sqlalchemy import select, func, text, String, bindparam, case, distinct
from server.app.db.dbmodels import (
    artifact, 
    artifactproperty, 
    context, 
    parentcontext, 
    attribution, 
    type_table, 
    association,
    execution,
    executionproperty,
    event,
)

async def register_server_details(db: AsyncSession, server_name: str, host_info: str):
    """
    Register server details in the database.
    """
    # Step 1: Check if the server is already registered
    query_check = text("""
        SELECT 1 FROM registered_servers WHERE host_info = :host_info
    """)
    result = await db.execute(query_check, {"host_info": host_info})
    # If a matching row exists, scalar() returns 1 (from SELECT 1). If not, it returns None
    exists = result.scalar()

    if exists:
        return {"message": "Server is already registered"}

    # Step 2: Insert new server
    query_insert = text("""
        INSERT INTO registered_servers (server_name, host_info)
        VALUES (:server_name, :host_info)
    """)
    await db.execute(query_insert, {"server_name": server_name, "host_info": host_info})
    await db.commit()

    return {"message": "Server registered successfully"}


async def get_registered_server_details(db: AsyncSession = Depends(get_db())):
    """
    Get all registered server details from the database.
    """
    query = text("""SELECT * FROM registered_servers""")
    result = await db.execute(query)
    return result.mappings().all()


async def get_sync_status(db: AsyncSession, server_name: str, host_info: str):
    """
    Get the sync status from the database.
    """
    query = text("""SELECT last_sync_time FROM registered_servers WHERE server_name = :server_name AND host_info = :host_info""")
    result = await db.execute(query, {"server_name": server_name, "host_info": host_info})
    return result.mappings().all()


async def update_sync_status(db: AsyncSession, current_utc_time: int, server_name: str, host_info: str):
    """
    Update the sync status in the database.
    """
    query = text("""
        UPDATE registered_servers
        SET last_sync_time = :current_utc_time
        WHERE server_name = :server_name AND host_info = :host_info
    """)
    await db.execute(query, {"current_utc_time": current_utc_time, "server_name": server_name, "host_info": host_info})
    await db.commit()  # Commit the transaction


async def fetch_artifacts(
    db: AsyncSession,   # Used to interact with the database
    pipeline_name: str, 
    artifact_type: str, 
    filter_value: str, 
    active_page: int = 1, 
    page_size: int = 5,  # Number of records per page
    sort_column: str = "name", 
    sort_order: str = "ASC"
):
    # Step 1: Get relevant contexts
    relevant_contexts_cte = select(
        parentcontext.c.context_id
    ).join(
        context, parentcontext.c.parent_context_id == context.c.id
    ).where(
        context.c.name == pipeline_name
    ).cte("relevant_contexts_cte")

    # Step 2: Fetch execution IDs based on pipeline name
    execution_ids_cte = (
        select(
            execution.c.id.label("execution_id")
        )
        .join(
            association, execution.c.id == association.c.execution_id
        )
        .where(
            association.c.context_id.in_(select(relevant_contexts_cte.c.context_id))
        )
        .cte("execution_ids_cte")
    )

    # Step 3: Based on execution ids list fetching equvalent artifact lists from event_path table
    artifact_ids_cte = (
        select(distinct(event.c.artifact_id).label("artifact_id"))
        .where(event.c.execution_id.in_(select(execution_ids_cte.c.execution_id)))
        .cte("artifact_ids_cte")
    )

    # Step 4: Aggregate artifact properties into JSON
    artifact_properties_agg_cte = (
        select(
            artifactproperty.c.artifact_id,
            func.json_agg(
                func.json_build_object(
                    "name", artifactproperty.c.name,
                    "value", func.coalesce(
                        artifactproperty.c.string_value,
                        func.cast(artifactproperty.c.bool_value, String),
                        func.cast(artifactproperty.c.double_value, String),
                        func.cast(artifactproperty.c.int_value, String),
                        func.cast(artifactproperty.c.byte_value, String),
                        func.cast(artifactproperty.c.proto_value, String),
                        text("NULL")
                    )
                )
            ).label("artifact_properties")
        )
        .where(artifactproperty.c.artifact_id.in_(select(artifact_ids_cte.c.artifact_id)))  # Filter by artifact IDs
        .group_by(artifactproperty.c.artifact_id)
        .cte("artifact_properties_agg_cte")
    )

    # Step 5: Aggregate execution type names per artifact
    artifact_execution_types_agg_cte = (
        select(
            event.c.artifact_id,
            func.string_agg(func.distinct(context.c.name), ', ').label("execution")
        )
        .join(execution, event.c.execution_id == execution.c.id)
        .join(association, execution.c.id == association.c.execution_id)
        .join(context, association.c.context_id == context.c.id)
        .where(event.c.artifact_id.in_(select(artifact_ids_cte.c.artifact_id)))
        .group_by(event.c.artifact_id)
        .cte("artifact_execution_types_agg_cte")
    )

    # Step 6: Base artifact metadata
    artifact_metadata_cte = (
        select(
            artifact.c.id.label('artifact_id'),
            artifact.c.name,
            artifact.c.uri,
            func.to_char(
                func.timezone('GMT', func.to_timestamp(artifact.c.create_time_since_epoch / 1000)),
                'Dy, DD Mon YYYY HH24:MI:SS GMT'
            ).label('create_time_since_epoch'),
            artifact.c.last_update_time_since_epoch
        )
        .join(type_table, artifact.c.type_id == type_table.c.id)
        .join(attribution, artifact.c.id == attribution.c.artifact_id)
        .join(context, attribution.c.context_id == context.c.id)
        .where(
            artifact.c.id.in_(select(artifact_ids_cte.c.artifact_id)),
            type_table.c.name == artifact_type
        )
        .cte("artifact_metadata_cte")
    )

    # Step 7: Query for fetching paginated data
    query = (
        select(
            artifact_metadata_cte.c.artifact_id,
            artifact_metadata_cte.c.name,
            artifact_execution_types_agg_cte.c.execution,
            artifact_metadata_cte.c.uri,
            artifact_metadata_cte.c.create_time_since_epoch,
            artifact_metadata_cte.c.last_update_time_since_epoch,
            artifact_properties_agg_cte.c.artifact_properties,
            func.count().over().label("total_records")  # Total records for pagination
        )
        .select_from(artifact_metadata_cte)
        .outerjoin(artifact_properties_agg_cte, artifact_metadata_cte.c.artifact_id == artifact_properties_agg_cte.c.artifact_id)
        .outerjoin(artifact_execution_types_agg_cte, artifact_metadata_cte.c.artifact_id == artifact_execution_types_agg_cte.c.artifact_id)
        .where(
            # Apply search filter to all columns of artifact_metadata_cte
            (artifact_metadata_cte.c.artifact_id.cast(String).ilike(f"%{filter_value}%")) |
            (artifact_metadata_cte.c.name.ilike(f"%{filter_value}%")) |
            (artifact_execution_types_agg_cte.c.execution.ilike(f"%{filter_value}%")) |
            (artifact_metadata_cte.c.uri.ilike(f"%{filter_value}%")) |
            (artifact_metadata_cte.c.create_time_since_epoch.cast(String).ilike(f"%{filter_value}%")) |
            (artifact_metadata_cte.c.last_update_time_since_epoch.cast(String).ilike(f"%{filter_value}%")) |
            # Apply search filter to artifact properties aggregation
            (artifact_properties_agg_cte.c.artifact_properties.cast(String).ilike(f"%{filter_value}%"))
        )
        .limit(page_size)  # Limit the number of records per page
        .offset((active_page - 1) * page_size)  # Offset for pagination
    )

    # Step 8: Apply sorting (order by the specified column and order)
    if sort_order.lower() == "desc":
        query = query.order_by(getattr(artifact_metadata_cte.c, sort_column).desc())
    else:
        query = query.order_by(getattr(artifact_metadata_cte.c, sort_column).asc())

    # Step 9: Execute the query
    result = await db.execute(query)
    rows = result.mappings().all()

    # Step 10: Extract total records and format results
    total_records = rows[0]["total_records"] if rows else 0
    return {
        "total_items": total_records,
        "items": [dict(row) for row in rows]
    }    


# acutal code
async def fetch_executions(
    db: AsyncSession, 
    pipeline_name: str, 
    filter_value: str, 
    active_page: int = 1, 
    record_per_page: int = 5,
    sort_column: str = "Context_Type", 
    sort_order: str = "ASC"
):
    # Step 1: Get relevant contexts
    relevant_contexts = select(
        parentcontext.c.context_id
    ).join(
        context, parentcontext.c.parent_context_id == context.c.id
    ).where(
        context.c.name == pipeline_name
    ).subquery("relevant_contexts")

    # Step 2: Aggregate execution properties into JSON
    execution_properties_agg = (
        select(
            executionproperty.c.execution_id,
            func.json_agg(
                func.json_build_object(
                    "name", executionproperty.c.name,
                    "value", func.coalesce(
                        executionproperty.c.string_value,
                        func.cast(executionproperty.c.bool_value, String), 
                        func.cast(executionproperty.c.double_value, String),
                        func.cast(executionproperty.c.int_value, String),
                        func.cast(executionproperty.c.byte_value, String),
                        func.cast(executionproperty.c.proto_value, String),
                        text("NULL") 
                    )
                )
            ).label("execution_properties")
        )
        .group_by(executionproperty.c.execution_id)
        .subquery()
    )

    # Step 3: Base data query (executions and associations)
    base_data = (
        select(
            execution.c.id.label("execution_id")
        )
        .join(
            type_table, execution.c.type_id == type_table.c.id
        )
        .join(
            association, execution.c.id == association.c.execution_id
        )
        .join(
            context, association.c.context_id == context.c.id
        )
        .where(
            association.c.context_id.in_(select(relevant_contexts.c.context_id))
        )
        .subquery("base_data")
    )

    # Step 4: Final query with execution properties aggregation
    query = (
        select(
            base_data.c.execution_id,
            execution_properties_agg.c.execution_properties,
            func.count().over().label("total_records")
        )
        .select_from(base_data)
        .outerjoin(
            execution_properties_agg, base_data.c.execution_id == execution_properties_agg.c.execution_id
        )
        .where(
            func.lower(execution_properties_agg.c.execution_properties.cast(String)).ilike(f"%{filter_value}%")  # Filter based on execution_properties using ILIKE
        )
        .limit(record_per_page)
        .offset((active_page - 1) * record_per_page)
    )

    # Step 5: Execute the query for paginated results
    result = await db.execute(query)
    rows = result.mappings().all()

    # Step 6: Extract total records from the first row (since it's constant across all rows)
    total_record = rows[0]["total_records"] if rows else 0

    # Return the results in the specified format
    return {
        "total_items": total_record,
        "items": [dict(row) for row in rows]
    }



import os
from dotenv import load_dotenv
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine, async_sessionmaker
from sqlalchemy.pool import NullPool  # For connection pooling (optional)

# Load .env variables
load_dotenv()

# Database configuration
DB_CONFIG = {
    "db_user": os.getenv("POSTGRES_USER"),
    "db_password": os.getenv("POSTGRES_PASSWORD"),
    "db": os.getenv("POSTGRES_DB"),
    "db_host_ip": os.getenv("MYIP"),
    "db_host_name": os.getenv("HOSTNAME"),
    "db_port": 5432, # Default PostgreSQL port
}

if DB_CONFIG['db_host_ip'] and DB_CONFIG['db_host_ip'] != "127.0.0.1":
    DB_CONFIG["db_host"] = DB_CONFIG['db_host_ip']
else:
    DB_CONFIG["db_host"] = DB_CONFIG['db_host_name']


DATABASE_URL = "postgresql+asyncpg://{0}:{1}@{2}:{3}/{4}".format(
        DB_CONFIG["db_user"], DB_CONFIG["db_password"], DB_CONFIG["db_host"],
        DB_CONFIG["db_port"], DB_CONFIG["db"]
        )


# Create an asynchronous engine with connection pooling
engine = create_async_engine(
    DATABASE_URL,
    echo=True,  # For debugging, log SQL statements
)

# Create a session maker
async_session = async_sessionmaker(
    bind=engine,
    class_=AsyncSession,
    expire_on_commit=False,
)

async def get_db():
    async with async_session() as session:
        yield session




from cmflib.cmfquery import CmfQuery
import pandas as pd
import json
import typing as t
from fastapi.concurrency import run_in_threadpool
from server.app.query_artifact_lineage_d3force import query_artifact_lineage_d3force
from server.app.query_list_of_executions import query_list_of_executions


#Converts sync functions to async
async def async_api(function_to_async, query: CmfQuery, *argv):
    return await run_in_threadpool(function_to_async, query, *argv)

async def get_model_data(query: CmfQuery, modelId: int):
    '''
      This function retrieves the necessary model data required for generating a model card.

      Arguments:
        modelId (int): The ID of the model for which data is required.

      Returns:
        This function returns a tuple of DataFrames containing the following:

        model_data_df (DataFrame): Metadata related to the model itself.
        model_exe_df (DataFrame): Metadata of the executions in which the specified modelId was an input or output.
        model_input_df (DataFrame): Metadata of input artifacts that led to the creation of the model.
        model_output_df (DataFrame): Metadata of artifacts that used the model as an input.
        The returned DataFrames provide comprehensive metadata for the specified model, aiding in the creation of detailed and accurate model cards.
    '''
    pd.set_option('display.max_columns', None)
    model_data_df = pd.DataFrame()
    model_exe_df = pd.DataFrame()
    model_input_df = pd.DataFrame()
    model_output_df = pd.DataFrame()

    # get name from id
    modelName = ""
    model_data_df = query.get_all_artifacts_by_ids_list([modelId])
    # if above dataframe is not empty, we have the dataframe for given modelId with full model related details
    if model_data_df.empty:
        return model_data_df, model_exe_df, model_input_df, model_output_df
    # However following check is done, in case, variable 'modelId' is not an ID for model artifact
    modelType = model_data_df['type'].tolist()[0]
    if not modelType == "Model":
        # making model_data_df empty
        model_data_df = pd.DataFrame()
        return model_data_df, model_exe_df, model_input_df, model_output_df


    # extracting modelName
    modelName = model_data_df['name'].tolist()[0]

    # model's executions data with props and custom props
    exe_df = query.get_all_executions_for_artifact(modelName)
    exe_ids = []
    if not exe_df.empty:
        exe_df.drop(columns=['execution_type_name', 'execution_name'], inplace=True)
        exe_ids = exe_df['execution_id'].tolist()


    if not exe_ids:
         return model_data_df, model_exe_df, model_input_df, model_output_df
    model_exe_df = query.get_all_executions_by_ids_list(exe_ids)
    model_exe_df.drop(columns=['Git_Start_Commit', 'Git_End_Commit'], inplace=True)

    in_art_ids =  []
    # input artifacts
    # it is usually not a good practice to use functions starting with _ outside of the file they are defined .. should i change??
    in_art_ids.extend(query._get_input_artifacts(exe_ids))
    if modelId in in_art_ids:
        in_art_ids.remove(modelId)
    model_input_df = query.get_all_artifacts_by_ids_list(in_art_ids)

    out_art_ids = []
    # output artifacts
    out_art_ids.extend(query._get_output_artifacts(exe_ids))
    if modelId in out_art_ids:
        out_art_ids.remove(modelId)
    model_output_df = query.get_all_artifacts_by_ids_list(out_art_ids)

    return model_data_df, model_exe_df, model_input_df, model_output_df


def get_executions(query: CmfQuery, pipeline_name, exe_ids) -> pd.DataFrame:
    '''
    Args:
     pipeline_name: name of the pipeline.
     exe_ids: list of execution ids.

    Returns:
     returns dataframe of executions using execution_ids.
    '''
    df = pd.DataFrame()
    executions = query.get_all_executions_by_ids_list(exe_ids)
    df = pd.concat([df, executions], sort=True, ignore_index=True)
    return df


def get_all_exe_ids(query: CmfQuery, pipeline_name: t.Optional[str] = None) -> t.Dict[str, pd.DataFrame]:
    '''
    Returns:
    returns a dictionary which has pipeline_name as key and dataframe which includes {id,Execution_uuid,Context_Type,Context_id} as value.
    '''
    execution_ids = {}
    executions = pd.DataFrame()    # df is emptied to store execution ids for next pipeline.
    if pipeline_name:
        executions = query.get_all_executions_in_pipeline(pipeline_name)
        if not executions.empty:
            execution_ids[pipeline_name] = executions[['id', 'Context_Type', 'Execution_uuid', 'Context_ID']]
        else:
            execution_ids[pipeline_name] = pd.DataFrame()
    else:
        names = query.get_pipeline_names()
        for name in names:
            executions = pd.DataFrame()    # df is emptied to store execution ids for next pipeline.
            executions = query.get_all_executions_in_pipeline(name)
            # check if df is empty return just pipeline_name: {}
            # if df is not empty return dictionary with pipeline_name as key
            # and df with id, context_type, uuid, context_ID as value.
            if not executions.empty:
                execution_ids[name] = executions[['id', 'Context_Type', 'Execution_uuid', 'Context_ID']]
            else:
                execution_ids[name] = pd.DataFrame()
    return execution_ids

def get_all_artifact_ids(query: CmfQuery, execution_ids, pipeline_name: t.Optional[str] = None) -> t.Dict[str, t.Dict[str, pd.DataFrame]]:
    # following is a dictionary of dictionaries

    # First level dictionary key is pipeline_name
    # First level dicitonary value is nested dictionary
    # Nested dictionary key is type i.e. Dataset, Model, etc.
    # Nested dictionary value is a pandas df with id and artifact name
    artifact_ids: t.Dict[str, t.Dict[str, pd.DataFrame]] = {}
    artifacts = pd.DataFrame()
    if pipeline_name:
        if not execution_ids.get(pipeline_name).empty:
            exe_ids = execution_ids[pipeline_name]['id'].tolist()
            artifacts = query.get_all_artifacts_for_executions(exe_ids)
            #acknowledging pipeline exist even if df is empty. 
            if artifacts.empty:
                artifact_ids[pipeline_name] = {}   # { pipeline_name: {empty dict} }
            else:
                artifact_ids[pipeline_name] = {}
                for art_type in artifacts['type']:
                    filtered_values = artifacts.loc[artifacts['type'] == art_type, ['id', 'name']]
                    artifact_ids[pipeline_name][art_type] = filtered_values
        # if execution_ids is empty then create dictionary with key as pipeline name
        # and value as empty df
        else:
            artifact_ids[pipeline_name] = {}
    else:
        names = query.get_pipeline_names()
        for name in names:
            if not execution_ids.get(name).empty:
                exe_ids = execution_ids[name]['id'].tolist()
                artifacts = query.get_all_artifacts_for_executions(exe_ids)
                #acknowledging pipeline exist even if df is empty. 
                if artifacts.empty:
                    artifact_ids[name] = {}   # { pipeline_name: {empty dict} }
                else:
                    artifact_ids[name] = {}
                    for art_type in artifacts['type']:
                        filtered_values = artifacts.loc[artifacts['type'] == art_type, ['id', 'name']]
                        artifact_ids[name][art_type] = filtered_values
            # if execution_ids is empty then create dictionary with key as pipeline name
            # and value as empty df
            else:
                artifact_ids[name] = {}
    return artifact_ids


def get_artifacts(query: CmfQuery, pipeline_name, art_type, artifact_ids):
    df = pd.DataFrame()
    if (query.get_pipeline_id(pipeline_name) != -1):
        df = query.get_all_artifacts_by_ids_list(artifact_ids)
        if len(df) == 0:
            return
        df = df.drop_duplicates()
        art_names = df['name'].tolist()
        name_dict = {}
        name_list = []
        exec_type_name_list = []
        exe_type_name = pd.DataFrame()
        for name in art_names:
            executions = query.get_all_executions_for_artifact(name)
            exe_type_name = pd.concat([exe_type_name, executions], ignore_index=True)
            execution_type_name = exe_type_name["execution_type_name"].drop_duplicates().tolist()
            execution_type_name = [str(element).split('"')[1] for element in execution_type_name]
            execution_type_name_str = ',\n '.join(map(str, execution_type_name))
            name_list.append(name)
            exec_type_name_list.append(execution_type_name_str)
        name_dict['name'] = name_list
        name_dict['execution_type_name'] = exec_type_name_list
        name_df = pd.DataFrame(name_dict)
        merged_df = df.merge(name_df, on='name', how='left')
        merged_df['name'] = merged_df['name'].apply(lambda x: x.split(':')[0] if ':' in x else x)
        merged_df = merged_df.loc[merged_df["type"] == art_type]
        result = merged_df.to_json(orient="records")
        tempout = json.loads(result)
        return tempout

def get_artifact_types(query: CmfQuery) -> t.List[str]:
    artifact_types = query.get_all_artifact_types()
    return artifact_types


def get_mlmd_from_server(query: CmfQuery, pipeline_name: t.Optional[str] = None, exec_uuid: t.Optional[str] = None, last_sync_time: t.Optional[str] = None, 
                         dict_of_exe_ids: t.Optional[dict] = None) -> t.Optional[str]:
    """
    Retrieves metadata from the server for a given pipeline and execution UUID if mentioned.

    Args:
        query (CmfQuery): The CmfQuery object.
        pipeline_name (str): The name of the pipeline.
        exec_uuid (str): The execution UUID.
        dict_of_exe_ids (dict): A dictionary containing execution IDs for pipelines.

    Returns:
        json_payload (str or None): The metadata in JSON format if found, "no_exec_uuid" 
        if the execution UUID is not found, or None if the pipeline name is not available.
    """
    json_payload = None
    flag=False
    if pipeline_name == None and not last_sync_time:
        # in case of first sync or second sync we don't know if there is one pipeline or multiple pipelines
        if last_sync_time is None:
            json_payload = query.extract_to_json(0)
    elif pipeline_name == None and last_sync_time:
        json_payload = query.extract_to_json(int(last_sync_time))
    else:
        if(pipeline_name is not None and query.get_pipeline_id(pipeline_name) != -1 and dict_of_exe_ids is not None):  # checks if pipeline name is available in mlmd
            if exec_uuid != None:
                dict_of_exe_ids = dict_of_exe_ids[pipeline_name]
                for index, row in dict_of_exe_ids.items():
                    exec_uuid_list = row['Execution_uuid'].split(",")
                    if exec_uuid in exec_uuid_list:
                        flag=True
                        break
                if not flag:
                    json_payload = "no_exec_uuid"
                    return json_payload
            json_payload = query.dumptojson(pipeline_name, exec_uuid)
    return json_payload


def get_lineage_data(
        query: CmfQuery, 
        pipeline_name, type,
        dict_of_art_ids,
        dict_of_exe_ids):
    """
    Retrieves lineage data based on the specified type.

    Parameters:
    query (CmfQuery): The CmfQuery object
    pipeline_name (str): The name of the pipeline.
    type (str): The type of lineage data to retrieve. Can be "Artifacts" or "Execution".
    dict_of_art_ids (dict): A dictionary of artifact IDs.
    dict_of_exe_ids (dict): A dictionary of execution IDs.

    Returns:
    dict or list: 
        - If type is "Artifacts", returns a dictionary with nodes and links for artifact lineage.
                lineage_data= {
                    nodes:[],
                    links:[]
                }
        - If type is "Execution", returns a list of execution types for the specified pipeline.
        - Otherwise, returns visualization data for artifact execution.
    """
    if type=="Artifacts":
        lineage_data = query_artifact_lineage_d3force(query, pipeline_name, dict_of_art_ids)
    elif type=="Execution":
        lineage_data = query_list_of_executions(pipeline_name, dict_of_exe_ids)
    else:
        pass
    return lineage_data



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import yaml
import uuid
import argparse
import typing as t
import pandas as pd

from cmflib import cmfquery, cmf
from cmflib.cli.command import CmdBase
from ml_metadata.metadata_store import metadata_store
from ml_metadata.proto import metadata_store_pb2 as mlpb
from cmflib.utils.helper_functions import fetch_cmf_config_path
from cmflib.cmf_exception_handling import MsgSuccess, FileNotFound

class CmdDVCIngest(CmdBase):
    
    def find_location(self, string, elements):
        for index, element in enumerate(elements):
            if string == element:
                return index
        return None
    
    def get_cmf_hierarchy(self, execution_lineage: str):
        """
            Parses the string and returns, pipeline name, context name and execution name
        """
        cmf_levels = execution_lineage.split(',')
        return cmf_levels[-1], cmf_levels[1], cmf_levels[0]

    def ingest_metadata(self, execution_lineage: str, metadata: dict, metawriter: cmf.Cmf, tracked: dict, command: str = "" ) :
        """
        Ingest the metadata into cmf
        args
            execution_lineage: format- execution id(if present)/execution file name/context/pipeline
            eg - demo_train.py,active_learning_training@2,active_learning -- without existing execution
                - 3,eval,active_learning -- with existing execution in cmf metadata file
            metadata: The parsed dictionary from dvc.lock file
            execution_exist : True if it exeist, False otherwise
            metawrite: cmf object 
        """
        pipeline_name, context_name, execution = self.get_cmf_hierarchy(execution_lineage)

        _ = metawriter.create_execution(
            str(context_name) + '_' + str(execution), 
            {}, 
            cmd = str(command),
            create_new_execution=False
            )
            
        output, config_file_path = fetch_cmf_config_path()
        artifact_repo = list(output.values())[0]

        for k, v in metadata.items():
            props={}
            if k == "deps":
                for dep in v:
                    if dep["path"] not in tracked:
                        metawriter.log_dataset(dep["path"], 'input')
                    else:
                        md5_value = dep["md5"] 
                        url = pipeline_name + ":" + artifact_repo + "/files/md5/" + md5_value[:2] + "/" + md5_value[2:]
                        props["url"] =  url
                        metawriter.log_dataset_with_version(dep["path"], md5_value, "input", props)
            if k == "outs":
                for out in v:
                    md5_value = out["md5"] 
                    url = pipeline_name + ":" + artifact_repo + "/files/md5/" + md5_value[:2] + "/" + md5_value[2:]
                    props["url"] =  url
                    metawriter.log_dataset_with_version(out["path"], md5_value, "output", props)
                    tracked[out["path"]] = True
        return tracked

    def run(self, live):
        """
            Ingest the metadata in dvc.lock file into CMF.
            If mlmd file exist with metadata from metrics logging or logging for other
            metadata not captured in dvc.lock file, pass that mlmd file as the input file.
            This code queries the file for existing pipelines, stages and executions and stores 
            as a dictionary. If the execution cmd in the stored dict, matches the execution command
            in the dvc.lock file, that execution is updated with additional metadata.
            If there is no prior execution captured, a new execution is created
        """
        uuid_ = str(uuid.uuid4())
        pipeline_name = ""
        cmd_exe: t.Dict[str, str] = {}
        query = cmfquery.CmfQuery(self.args.file_name)
        pipelines: t.List[str] = query.get_pipeline_names()

        # Query mlmd to get all the executions and its commands
        for pipeline in pipelines:
            pipeline_name = pipeline
            stages: t.List[str] = query.get_pipeline_stages(pipeline)
            for stage in stages:
                exe_df: pd.DataFrame = query.get_all_executions_in_stage(stage)
                """
                Parse all the executions in a stage
                eg- exe_step = ['demo_eval.py', '--trained_model', 'data/model-1', '--enable_df', 'True', '--round', '1']
                """
                for index, row in exe_df.iterrows():
                    exe_step = row['Execution']
                    '''
                    if already same execution command has been captured previously use the latest
                    execution id to associate the new metadata
                    '''
                    existing = cmd_exe.get(exe_step)
                    if existing is None:
                        cmd_exe[exe_step] = f"{row['id']},{stage},{pipeline}"
                    else:
                        if row['id'] > int(existing.split(',')[0]):
                            cmd_exe[exe_step] = f"{row['id']},{stage},{pipeline}"

        # Parse the dvc.lock file.
        # pipeline_dict stores pipeline stages with the following hierarchy:
        # {
        #     "<stage_name>": {
        #         "<index>": {
        #             "cmd": List[str],        # List of command parts as strings
        #             "deps": List[Dict[str, Any]],  # List of dependency dictionaries
        #             "outs": List[Dict[str, Any]]   # List of output dictionaries
        #         }
        #     }
        # }
        pipeline_dict: t.Dict[str, t.Dict[str, t.Dict[str, t.Union[t.List[str], t.List[t.Dict[str, t.Any]]]]]] = {}
        
        # check whether dvc.lock file present in cwd
        if not os.path.exists("dvc.lock"):
            raise FileNotFound("dvc.lock", os.getcwd())

        with open("dvc.lock", 'r') as f:
            valuesYaml = yaml.load(f, Loader=yaml.FullLoader)

        for k in valuesYaml['stages']:
            pipeline_dict[k] = {}
            commands=[]
            deps = []
            outs = []
            k_dict = {}
            i = 0
            
            for kk in valuesYaml['stages'][k]:
                if kk == 'cmd':
                    cmd_list = valuesYaml['stages'][k][kk].split()
                    commands.append(cmd_list)
                    k_dict['cmd'] = cmd_list
                    i = i + 1
                if kk == 'deps':
                    deps = valuesYaml['stages'][k][kk]
                    k_dict['deps'] = deps
                if kk == 'outs':
                    outs = valuesYaml['stages'][k][kk]
                    k_dict['outs'] = outs

            pipeline_dict[k][str(i)] = k_dict

        # Create a unique Pipeline name if there is no mlmd file
        pipeline_name = "Pipeline"+"-"+str(uuid_) if not pipeline_name else pipeline_name
        metawriter = cmf.Cmf(filepath = "mlmd", pipeline_name=pipeline_name, graph=True)

        # Parse the dvc.lock dictionary and get the command section
        tracked = {} #Used to keep a record of files tracked by outs and therefore not needed to be tracked in deps
        for k, v in pipeline_dict.items():
            for kk, vv in v.items():
                for kkk, vvv in vv.items():
                    if kkk == 'cmd':
                        """
                        Key eg - cmd
                        Value eg - ['python3', 'demo.py', '--enable_df', 'True']
                        cmd_exe eg - {"['demo_eval.py', '--trained_model', 'data/model-3', '--enable_df', 'True', '--round', '3']":
                        '3,eval,active_learning', 
                        "['demo_eval.py', '--trained_model', 'data/model-2', '--enable_df', 'True', '--round', '2']": '2,eval,active_learning',
                        "['demo_eval.py', '--trained_model', 'data/model-1', '--enable_df', 'True', '--round', '1']": '1,eval,active_learning'}
                        In the next line pop out the python
                        if the pipeline_dict command is already there in the cmd_exe dict got from parsing the mlmd pop that cmd out 
                        and use the stored lineage from the mlmd
                        """
                        vvv.pop(0)
                        pos = self.find_location('--execution_name', vvv)
                        if pos:
                            execution_name = vvv[pos+1]
                        else:
                            execution_name = uuid_
                            
                        context_name = k
                        lineage = execution_name+","+context_name+","+ pipeline_name
                        # Cast vvv to a list of strings to ensure join works correctly.
                        # join the list to single string
                        cmd_str = ' '.join(t.cast(t.List[str], vvv))
                        cmd = cmd_exe.get(cmd_str, None)
                        _ = metawriter.create_context(pipeline_stage=context_name)
                        tracked = self.ingest_metadata(lineage, vv, metawriter, tracked, cmd_str)
        metawriter.log_dvc_lock("dvc.lock")
        return MsgSuccess(msg_str="Command has completed running..")


def add_parser(subparsers, parent_parser):
    HELP = "Ingests metadata from the dvc.lock file into the CMF."

    parser = subparsers.add_parser(
        "ingest",
        parents=[parent_parser],
        description="Ingests metadata from the dvc.lock file into the CMF. If an existing MLMD file is provided, it merges and updates execution metadata based on matching commands, or creates new executions if none exist.",
        help=HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "-f", 
        "--file_name", 
        type=str,
        default="mlmd",
        help="Specify input mlmd file name. (default: mlmd)",
        metavar="<file_name>"
    )

    parser.set_defaults(func=CmdDVCIngest)



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import os
import re
import yaml
import argparse

from cmflib import cmfquery
from cmflib.cli.command import CmdBase
from cmflib.cli.utils import check_minio_server
from cmflib.utils.helper_functions import generate_osdf_token, fetch_cmf_config_path
from cmflib.dvc_wrapper import dvc_push, dvc_add_attribute
from cmflib.utils.cmf_config import CmfConfig
from cmflib.cmf_exception_handling import (
    PipelineNotFound, Minios3ServerInactive, 
    FileNotFound, 
    ExecutionsNotFound,
    ArtifactPushSuccess, 
    MissingArgument, 
    DuplicateArgumentNotAllowed
)

class CmdArtifactPush(CmdBase):
    def run(self, live):
        dvc_config_op, config_file_path = fetch_cmf_config_path()
        
        cmd_args = {
            "file_name": self.args.file_name,
            "pipeline_name": self.args.pipeline_name, 
            "jobs": self.args.jobs,
        }
        for arg_name, arg_value in cmd_args.items():
            if arg_value:
                if arg_value[0] == "":
                    raise MissingArgument(arg_name)
                elif len(arg_value) > 1:
                    raise DuplicateArgumentNotAllowed(arg_name,("-"+arg_name[0]))

        out_msg = check_minio_server(dvc_config_op)
        if dvc_config_op["core.remote"] == "minio" and out_msg != "SUCCESS":
            raise Minios3ServerInactive()
        
        # If user has not specified the number of jobs or jobs is not a digit, set it to 4 * cpu_count()
        num_jobs = int(self.args.jobs[0]) if self.args.jobs and self.args.jobs[0].isdigit() else 4 * os.cpu_count()
        
        if dvc_config_op["core.remote"] == "osdf":
            cmf_config={}
            cmf_config=CmfConfig.read_config(config_file_path)
            #print("key_id="+cmf_config["osdf-key_id"])
            dynamic_password = generate_osdf_token(cmf_config["osdf-key_id"],cmf_config["osdf-key_path"],cmf_config["osdf-key_issuer"])
            #print("Dynamic Password"+dynamic_password)
            dvc_add_attribute(dvc_config_op["core.remote"],"password",dynamic_password)
            #The Push URL will be something like: https://<Path>/files/md5/[First Two of MD5 Hash]
            result = dvc_push(num_jobs=num_jobs)
            #print(result)
            return result

        # Default path of mlmd file
        current_directory = os.getcwd()
        if not self.args.file_name:         # If self.args.file_name is None or an empty list ([]). 
            mlmd_file_name = "./mlmd"       # Default path for mlmd file name.
        else:
            mlmd_file_name = self.args.file_name[0].strip()
            if mlmd_file_name == "mlmd":
                mlmd_file_name = "./mlmd"
        current_directory = os.path.dirname(mlmd_file_name)
        if not os.path.exists(mlmd_file_name):
            raise FileNotFound(mlmd_file_name, current_directory)
        
        # creating cmfquery object
        query = cmfquery.CmfQuery(mlmd_file_name)
        
        pipeline_name = self.args.pipeline_name[0]
        # Put a check to see whether pipline exists or not
        if not pipeline_name in query.get_pipeline_names():
            raise PipelineNotFound(pipeline_name)

        stages = query.get_pipeline_stages(pipeline_name)
        executions = []
        identifiers = []

        for stage in stages:
            # getting all executions for stages
            executions = query.get_all_executions_in_stage(stage)
            # check if stage has executions
            if len(executions) > 0:
                # converting it to dictionary
                dict_executions = executions.to_dict("dict")
                for id in dict_executions["id"].values():
                    identifiers.append(id)
            else:
                print("No Executions found for " + stage + " stage.")

        names = []
        if len(identifiers) == 0:  # check if there are no executions
            raise ExecutionsNotFound()
        for identifier in identifiers:
            artifacts = query.get_all_artifacts_for_execution(
                 identifier
            )  # getting all artifacts with id
            # dropping artifact with type 'metrics' as metrics doesn't have physical file
            if not artifacts.empty:
                artifacts = artifacts[artifacts['type'] != 'Metrics']
                # adding .dvc at the end of every file as it is needed for pull
                artifacts['name'] = artifacts['name'].apply(lambda name: f"{name.split(':')[0]}.dvc")
                names.extend(artifacts['name'].tolist())
        final_list = set()
        for file in set(names):
            # checking if the .dvc exists
            if os.path.exists(file):
                final_list.add(file)
            # checking if the .dvc exists in user's project working directory
            elif os.path.isabs(file):
                file = re.split("/",file)[-1]
                file = os.path.join(os.getcwd(), file)
                if os.path.exists(file):
                    final_list.add(file)
            else:
                # in case of dvc_ingest_command
                # fetching remaining artifacts from dvc.lock file
                if os.path.exists("dvc.lock"):
                    with open("dvc.lock", "r") as f:
                        str_data = f.read()
                    data = yaml.safe_load(str_data)
                    # Traverse all stages and collect all 'path' keys from both 'deps' and 'outs'
                    for stage in data.get('stages', {}).values():
                        for section in ['deps', 'outs']:
                            for item in stage.get(section, []):
                                if isinstance(item, dict) and 'path' in item:
                                    final_list.add(item['path'])
        #print("file_set = ", final_list)
        result = dvc_push(num_jobs, list(final_list))
        return ArtifactPushSuccess(result)
    
def add_parser(subparsers, parent_parser):
    HELP = "Push artifacts to the user configured artifact repo."

    parser = subparsers.add_parser(
        "push",
        parents=[parent_parser],
        description="Push artifacts to the user configured artifact repo.",
        help=HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "-p",
        "--pipeline_name",
        required=True,
        action="append",
        help="Specify Pipeline name.",
        metavar="<pipeline_name>",
    )

    parser.add_argument(
        "-f", 
        "--file_name", 
        action="append",
        help="Specify input metadata file name.",
        metavar="<file_name>"
    )

    parser.add_argument(
        "-j",
        "--jobs",
        action="append",
        help="Number of parallel jobs for uploading artifacts to remote storage. Default is 4 * cpu_count(). Increasing jobs may speed up uploads but will use more resources.",
        metavar="<jobs>"
    )

    parser.set_defaults(func=CmdArtifactPush)



from cmflib.cmfquery import CmfQuery
from collections import deque, defaultdict
from typing import List, Dict, Any
from server.app.utils import modify_arti_name

def query_artifact_lineage_d3tree(query: CmfQuery, pipeline_name: str, dict_of_art_ids: Dict) -> List[List[Dict[str, Any]]]:
    env_list = []
    id_name = {}
    child_parent_artifact_id: Dict[int, List[int]] = {}
    for type_, df in dict_of_art_ids[pipeline_name].items():
        if type_ == "Environment":
            env_list = list(df["id"])
        for index, row in df.iterrows():
            #creating a dictionary of id and artifact name {id:artifact name}
            artifact_id = row['id']  # This will be an integer
            if artifact_id in env_list:
                continue
            id_name[artifact_id] = modify_arti_name(row["name"], type_)
            one_hop_parent_artifacts = query.get_one_hop_parent_artifacts_with_id(artifact_id)  # get immediate artifacts     
            child_parent_artifact_id[artifact_id] = []      # assign empty dict for artifact with no parent artifact
            if not one_hop_parent_artifacts.empty:        # if artifact have parent artifacts    
                parents_list =  list(one_hop_parent_artifacts["id"])
                final_parents_list = list(set(parents_list) - set(env_list))
                #child_parent_artifact_id[artifact_id] = list(one_hop_parent_artifacts["id"])
                child_parent_artifact_id[artifact_id] = final_parents_list
    data_organized: List[List[Dict[str, Any]]] = topological_sort(child_parent_artifact_id, id_name)
    return data_organized

def topological_sort(input_data, artifact_name_id_dict) -> List[List[Dict[str, Any]]]:
    # Initialize in-degree of all nodes to 0
    in_degree = {node: 0 for node in input_data}
    # Initialize adjacency list
    adj_list = defaultdict(list)

    # Fill the adjacency list and in-degree dictionary
    for node, dependencies in input_data.items():
        for dep in dependencies:
            adj_list[dep].append(node)
            in_degree[node] += 1

    # Queue for nodes with in-degree 0
    zero_in_degree_queue = deque([node for node, degree in in_degree.items() if degree == 0])
    topo_sorted_nodes = []

    while zero_in_degree_queue:
        current_node = zero_in_degree_queue.popleft()
        topo_sorted_nodes.append(current_node)
        for neighbor in adj_list[current_node]:
            in_degree[neighbor] -= 1
            if in_degree[neighbor] == 0:
                zero_in_degree_queue.append(neighbor)
    # Transform sorted nodes into the required output format
    parent_dict = defaultdict(list)
    for id_val in topo_sorted_nodes:   # topo_sorted_nodes = ['1','2','3','4']
        if id_val in input_data:       # input_data = {"child_id":[parents_id]}, for example {"4":['3','7','9']}
            parents = tuple(sorted(input_data[id_val]))
            parent_dict[parents].append({'id': artifact_name_id_dict[id_val],'parents': [artifact_name_id_dict[parent] for parent in input_data[id_val]]})
    output_data= list(parent_dict.values()) 
    return output_data



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import typing as t
from minio import Minio
from minio.error import S3Error
from cmflib.cmf_exception_handling import BucketNotFound

class MinioArtifacts:

    def __init__(self, dvc_config_op):
        """
        Initialize the MinioArtifacts class with minios3 repo credentials.
        dvc_config_op["remote.minio.endpointurl"] = http://XX.XX.XX.XX:9000
        Args:
            dvc_config_op (dict): Dictionary containing local url (remote.local.url).
        """
        
        self.endpoint = dvc_config_op["remote.minio.endpointurl"].split("http://")[1]
        self.access_key = dvc_config_op["remote.minio.access_key_id"]
        self.secret_key = dvc_config_op["remote.minio.secret_access_key"]
        self.client = Minio(
                self.endpoint, access_key=self.access_key, secret_key=self.secret_key, secure=False
            )


    def download_file(
        self, 
        current_directory: str,
        bucket_name: str,
        object_name: str,
        download_loc: str,
        ):
        """
        Download a single file from an S3 bucket.

        Args:
            current_directory (str): The current working directory.
            bucket_name (str): Name of the minioS3 bucket.
            object_name (str): Key (path) of the file in the minios3 repo.
            download_loc (str): Local path where the file should be downloaded.

        Returns:
            tuple: (object_name, download_loc, status) where status indicates success (True) or failure (False).
        """
        try:
            found = self.client.bucket_exists(bucket_name)

            #check if minio bucket exists
            if not found:   
                raise BucketNotFound(bucket_name)

            #Download file
            response = self.client.fget_object(bucket_name, object_name, download_loc)

            # Check if the response indicates success.
            if response:
                return object_name, download_loc, True
            return object_name, download_loc, False
        except S3Error as exception:
            print(exception)
            return object_name, download_loc, False
        except Exception as e:
            return object_name, download_loc, False


    def download_directory(
        self,
        current_directory: str,
        bucket_name: str,
        object_name: str,
        download_loc: str,
    ):
        """
        Download a directory from an minios3 repo using its .dir metadata object.

        Args:
            current_directory (str): The current working directory .
            bucket_name (str): Name of the minios3 bucket.
            object_name (str): Key (path) of the .dir object in the minios3 bucket.
            download_loc (str): Local directory path where the directory should be downloaded.

        Returns:
            tuple: (total_files_in_directory, files_downloaded, status) where status indicates success (True) or failure (False).
        """

        found = self.client.bucket_exists(bucket_name)
        if not found:   #check if minio bucket exists
            raise BucketNotFound(bucket_name)

        """"
        if object_name ends with .dir - it is a directory.
        we download .dir object with 'temp_dir' and remove 
        this after all the files from this .dir object is downloaded.
        """
                
        # in case of .dir, download_loc is a absolute path for a folder
        os.makedirs(download_loc, mode=0o777, exist_ok=True)

        total_files_in_directory = 0
        files_downloaded = 0

        # Temporary file to download the .dir metadata object.
        temp_dir = f"{download_loc}/temp_dir"
        try:
            # Download the .dir file containing metadata about tracked files.
            response = self.client.fget_object(bucket_name, object_name, temp_dir)

            with open(temp_dir, 'r') as file:
                tracked_files = eval(file.read())

            # removing temp_dir
            if os.path.exists(temp_dir):
                os.remove(temp_dir)

            """
            object_name =  files/md5/c9/d8fdacc0d942cf8d7d95b6301cfb97.dir
            contains the path of the .dir on the artifact repo
            we need to remove the hash of the .dir from the object_name
            which will leave us with the artifact repo path
            """
            repo_path_list: t.List[str] = object_name.split("/")
            repo_path_list = repo_path_list[:len(repo_path_list)-2]
            repo_path = "/".join(repo_path_list)

            obj=True
            for file_info in tracked_files:
                total_files_in_directory += 1
                relpath = file_info['relpath']
                md5_val = file_info['md5']
                # download_loc =  /home/sharvark/datatslice/example-get-started/test/artifacts/raw_data
                # md5_val = a237457aa730c396e5acdbc5a64c8453
                # we need a2/37457aa730c396e5acdbc5a64c8453
                formatted_md5 = md5_val[:2] + '/' + md5_val[2:]
                temp_download_loc = f"{download_loc}/{relpath}"
                temp_object_name = f"{repo_path}/{formatted_md5}"
                try:
                    obj = self.client.fget_object(bucket_name, temp_object_name, temp_download_loc)
                    if obj:
                        files_downloaded +=1
                        print(f"object {temp_object_name} downloaded at {temp_download_loc}.")
                    else:
                        print(f"object {temp_object_name} is not downloaded.")
                except Exception as e:
                    print(f"object {temp_object_name} is not downloaded.")
                    
            # total_files - files_downloaded gives us the number of files which are failed to download
            if (total_files_in_directory - files_downloaded) == 0:   
                return total_files_in_directory, files_downloaded, True
            return total_files_in_directory, files_downloaded, False  
        except S3Error as exception:
            print(exception)
            total_files_in_directory = 1 
            return total_files_in_directory, files_downloaded, False
        except Exception as e:
            print(f"object {object_name} is not downloaded.")
            # need to improve this  
            # We usually don't count .dir as a file while counting total_files_in_directory.
            # However, here we failed to download the .dir folder itself. So we need to make 
            # total_files_in_directory = 1, because  ..............
            total_files_in_directory = 1 
            return total_files_in_directory, files_downloaded, False
    
        # sometimes we get TypeError as an execption, however investiagtion for the exact scenarios is pending



###
# Copyright (2024) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import argparse

from cmflib import cmfquery
from cmflib.cli.command import CmdBase
from cmflib.cmf_exception_handling import (
    FileNotFound, 
    DuplicateArgumentNotAllowed, 
    MissingArgument, 
    MsgSuccess
)

class CmdPipelineList(CmdBase):
    def run(self, live):        
        current_directory = os.getcwd()
        if not self.args.file_name:         # If self.args.file_name is None or an empty list ([]). 
            mlmd_file_name = "./mlmd"       # Default path for mlmd file name.
        elif len(self.args.file_name) > 1:  # If the user provided more than one file name. 
            raise DuplicateArgumentNotAllowed("file_name", "-f")
        elif not self.args.file_name[0]:    # self.args.file_name[0] is an empty string ("").  
            raise MissingArgument("file name")
        else:
            mlmd_file_name = self.args.file_name[0].strip()
            if mlmd_file_name == "mlmd":
                mlmd_file_name = "./mlmd"
        
        current_directory = os.path.dirname(mlmd_file_name)
        if not os.path.exists(mlmd_file_name):
            raise FileNotFound(mlmd_file_name, current_directory)

        # Creating cmfquery object.
        query = cmfquery.CmfQuery(mlmd_file_name)
        
        return MsgSuccess(msg_list = [pipeline.name for pipeline in query._get_pipelines()])


def add_parser(subparsers, parent_parser):
    PIPELINE_LIST_HELP = "Display a list of pipeline name(s) from the available input metadata file."

    parser = subparsers.add_parser(
        "list",
        parents=[parent_parser],
        description="Display a list of pipeline name(s) from the available input metadata file.",
        help=PIPELINE_LIST_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    
    parser.add_argument(
        "-f", 
        "--file_name", 
        action="append",
        help="Specify input metadata file name.",
        metavar="<file_name>",
    )

    parser.set_defaults(func=CmdPipelineList)



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import requests
#import urllib3
#urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
import hashlib
import time

from urllib.parse import urlparse

def generate_cached_url(url, cache):
    #This takes host URL as supplied from MLMD records and generates cached URL=cache_path + path
    #Example Input: https://sdsc-origin.nationalresearchplatform.org:8443/nrp/fdp/23/6d9502e0283d91f689d7038b8508a2
    #Example Output: https://osdf-director.osg-htc.org/nrp/fdp/23/6d9502e0283d91f689d7038b8508a2  
    #The assumption is that url obtained from MLMD is more accurate. So we use the path from this URL and append it to cache path
    #but we clean up the cache path to only its scheme + netloc: https://osdf-director.osg-htc.org
    parsed_url = urlparse(url)
    parsed_cache_url= urlparse(cache) 
    cached_url= parsed_cache_url.scheme + "://" + parsed_cache_url.netloc + parsed_url.path
    return cached_url 

def calculate_md5_from_file(file_path, chunk_size=8192):
    md5 = hashlib.md5()
    try:
        with open(file_path, 'rb') as f:
            while chunk := f.read(chunk_size):
                md5.update(chunk)
    except Exception as e:
        print(f"An error occurred while reading the file: {e}")
        return None
    return md5.hexdigest()

def download_and_verify_file(host, headers, remote_file_path, local_path, artifact_hash, timeout):
    print(f"Fetching artifact={local_path}, surl={host} to {remote_file_path}")
    data= None
    try:
        response = requests.get(host, headers=headers, timeout=timeout, verify=True)  # This should be made True. otherwise this will produce Insecure SSL Warning
        if response.status_code == 200 and response.content:
            data = response.content
        else:
            return False, "No data received from the server."
            #pass
    except requests.exceptions.Timeout:
        return False, "The request timed out."
        #pass
    except Exception as exception:
        return False, str(exception)

    if data is not None:
        try: 
            with open(remote_file_path, 'wb') as file:
                file.write(data)
            if os.path.exists(remote_file_path) and os.path.getsize(remote_file_path) > 0:
                # Calculate MD5 hash of the downloaded file
                start_time = time.time()
                md5_hash = calculate_md5_from_file(remote_file_path)
                end_time = time.time()
                time_taken = end_time - start_time
                if md5_hash:
                    #print(f"MD5 hash of the downloaded file is: {md5_hash}")
                    #print(f"Time taken to calculate MD5 hash: {time_taken:.2f} seconds")
                    if artifact_hash == md5_hash:
                        #print("MD5 hash of the downloaded file matches the hash in MLMD records.")
                        stmt = f"object {local_path} downloaded at {remote_file_path} in {time_taken:.2f} seconds and matches MLMD records."
                        success=True
                    else:
                        #print("Error: MD5 hash of the downloaded file does not match the hash in MLMD records.")
                        stmt = f"object {local_path} downloaded at {remote_file_path} in {time_taken:.2f} seconds and does NOT match MLMD records."
                        success=False
                    return success, stmt
                else:
                    print("Failed to calculate MD5 hash of the downloaded file.")
        except Exception as e:
            print(f"An error occurred while writing to the file: {e}")
            return False, f"An error occurred while writing to the file: {e}"
    
    return False, "Data is None."


class OSDFremoteArtifacts:
    def download_artifacts(
        self,
        dvc_config_op,
        host: str, #s_url 
        cache: str, #cache_path from cmfconfig
        current_directory: str, #current_directory where cmf artifact pull is executed
        remote_file_path: str, # download_loc of the artifact 
        local_path: str, #name of the artifact
        artifact_hash: str, #hash of the artifact from MLMD records
    ):
        #print(f"Configured Host from MLMD record={host}. User configured cache redirector={cache}")
        output = ""
        remote_repo = dvc_config_op["remote.osdf.url"]
        user = "nobody"
        dynamic_password = dvc_config_op["remote.osdf.password"]
        custom_auth_header = dvc_config_op["remote.osdf.custom_auth_header"]
        #print(f"dynamic password from download_artifacts={dynamic_password}")
        #print(f"Fetching artifact={local_path}, surl={host} to {remote_file_path} when this has been called at {current_directory}")

        # Prepare directories and file paths
        headers={dvc_config_op["remote.osdf.custom_auth_header"]: dvc_config_op["remote.osdf.password"]}
        temp = local_path.split("/")
        temp.pop()
        dir_path = "/".join(temp)
        dir_to_create = os.path.join(current_directory, dir_path)
        os.makedirs(
            dir_to_create, mode=0o777, exist_ok=True
        )  # creates subfolders needed as per artifacts folder structure
        local_file_path = os.path.join(current_directory, local_path)
        local_file_path = os.path.abspath(local_file_path)
        
        #Cache can be Blank. If so, fetch from Origin
        if cache == "":
            #Fetch from Origin
            success, result = download_and_verify_file(host, headers, remote_file_path, local_file_path, artifact_hash, timeout=10)
            if success:
                #print(result)
                return success, result
            #print(f"Failed to download and verify file: {result}")
            return success, f"Failed to download and verify file: {result}"
        else:
            #Generate Cached path for artifact
            cached_s_url=generate_cached_url(host,cache)
            #Try to fetch from cache first
            success, cached_result = download_and_verify_file(cached_s_url, headers, remote_file_path, local_path, artifact_hash,timeout=5)
            if success:
                #print(cached_result)
                return success, cached_result
            else:
                print(f"Failed to download and verify file from cache: {cached_result}")
                print(f"Trying Origin at {host}")
                #Fetch from Origin 
                success, origin_result = download_and_verify_file(host, headers, remote_file_path, local_path, artifact_hash, timeout=10)
                if success: 
                    #print(origin_result)
                    return success, origin_result
                #print(f"Failed to download and verify file: {result}")
                return success, f"Failed to download and verify file: {origin_result}"
        

        


from sqlalchemy import(
    Table,
    Column,
    Integer,
    BigInteger,
    Text,
    String,
    Boolean,
    Double,
    LargeBinary,
    Index,
    UniqueConstraint,
    MetaData,
    SmallInteger
)

metadata = MetaData()

artifact = Table(
    "artifact", metadata,
    Column("id", Integer, primary_key=True, nullable=False),
    Column("type_id", Integer, nullable=False),
    Column("uri", Text),
    Column("state", Integer),
    Column("name", String(255)),
    Column("external_id", String(255), unique=True),
    Column("create_time_since_epoch", BigInteger, nullable=False),
    Column("last_update_time_since_epoch", BigInteger, nullable=False),
    
    # Indexes
    Index("idx_artifact_create_time_since_epoch", "create_time_since_epoch"),
    Index("idx_artifact_external_id", "external_id"),
    Index("idx_artifact_last_update_time_since_epoch", "last_update_time_since_epoch"),
    Index("idx_artifact_uri", "uri"),
    
    # Constraints
    UniqueConstraint("type_id", "name", name="uniqueartifacttypename")
)


artifactproperty = Table(
    "artifactproperty", metadata,
    Column("artifact_id", Integer, nullable=False),
    Column("name", String(255), nullable=False),
    Column("is_custom_property", Boolean, nullable=False),
    Column("int_value", Integer),
    Column("double_value", Double),
    Column("string_value", Text),
    Column("byte_value", LargeBinary),
    Column("proto_value", LargeBinary),
    Column("bool_value", Boolean),
    
    # Primary Key
    UniqueConstraint("artifact_id", "name", "is_custom_property", name="artifactproperty_pkey"),
    
    # Indexes
    Index("idx_artifact_property_double", "name", "is_custom_property", "double_value"),
    Index("idx_artifact_property_int", "name", "is_custom_property", "int_value"),
    Index("idx_artifact_property_string", "name", "is_custom_property", "string_value")
)


context = Table(
    "context", metadata,
    Column("id", Integer, primary_key=True, nullable=False),
    Column("type_id", Integer, nullable=False),
    Column("name", String(255), nullable=False),
    Column("external_id", String(255), unique=True),
    Column("create_time_since_epoch", BigInteger, nullable=False),
    Column("last_update_time_since_epoch", BigInteger, nullable=False),

    # Unique Constraints
    UniqueConstraint("type_id", "name", name="context_type_id_name_key"),

    # Indexes
    Index("idx_context_create_time_since_epoch", "create_time_since_epoch"),
    Index("idx_context_external_id", "external_id"),
    Index("idx_context_last_update_time_since_epoch", "last_update_time_since_epoch")
)

parentcontext = Table(
    "parentcontext", metadata, 
    Column("context_id", Integer, nullable=False),
    Column("parent_context_id", Integer, nullable=False),

    # Primary Key
    UniqueConstraint("context_id", "parent_context_id", name="parentcontext_pkey"),

    # Indexes
    Index("idx_parentcontext_parent_context_id", "parent_context_id")
)

type_table = Table(
    "type", metadata,
    Column("id", Integer, primary_key=True, nullable=False),
    Column("name", String(255), nullable=False),
    Column("version", String(255)),
    Column("type_kind", SmallInteger, nullable=False),
    Column("description", Text),
    Column("input_type", Text),
    Column("output_type", Text),
    Column("external_id", String(255), unique=True),

    # Indexes 
    Index("idx_type_external_id", "external_id"),
    Index("idx_type_name", "name"),

    # Constraints
    UniqueConstraint("external_id", name="type_external_id_key")
)


attribution = Table(
    "attribution", metadata, 
    Column("id", Integer, primary_key=True, nullable=False),
    Column("context_id", Integer, nullable=False),
    Column("artifact_id", Integer, nullable=False),

    # Unique Constraint
    UniqueConstraint("context_id", "artifact_id", name="attribution_context_id_artifact_id_key") 
)


execution = Table(
    "execution", metadata,
    Column("id", Integer, primary_key=True, nullable=False),
    Column("type_id", Integer, nullable=False),
    Column("last_known_state", Integer),
    Column("name", String(255)),
    Column("external_id", String(255), unique=True),
    Column("create_time_since_epoch", BigInteger, nullable=False),
    Column("last_update_time_since_epoch", BigInteger, nullable=False),
    
    # Indexes
    Index("idx_execution_create_time_since_epoch", "create_time_since_epoch"),
    Index("idx_execution_external_id", "external_id"),
    Index("idx_execution_last_update_time_since_epoch", "last_update_time_since_epoch"),
    
    # Constraints
    UniqueConstraint("type_id", "name", name="uniqueexecutiontypename")
    # UniqueConstraint("external_id", name="execution_external_id_key")
)


executionproperty = Table(
    "executionproperty", metadata,
    Column("execution_id", Integer, nullable=False),
    Column("name", String(255), nullable=False),
    Column("is_custom_property", Boolean, nullable=False),
    Column("int_value", Integer),
    Column("double_value", Double),
    Column("string_value", Text),
    Column("byte_value", LargeBinary),
    Column("proto_value", LargeBinary),
    Column("bool_value", Boolean),
    
    # Primary Key
    UniqueConstraint("execution_id", "name", "is_custom_property", name="executionproperty_pkey"),
    
    # Indexes
    Index("idx_execution_property_double", "name", "is_custom_property", "double_value"),
    Index("idx_execution_property_int", "name", "is_custom_property", "int_value"),
    Index("idx_execution_property_string", "name", "is_custom_property", "string_value")
)


association = Table(
    "association", metadata, 
    Column("id", Integer, primary_key=True, nullable=False),
    Column("context_id", Integer, nullable=False),
    Column("execution_id", Integer, nullable=False),

    # Unique Constraint
    UniqueConstraint("context_id", "execution_id", name="association_context_id_execution_id_key") 
)


event = Table(
    "event", metadata, 
    Column("id", Integer, primary_key=True, nullable=False),
    Column("artifact_id", Integer, nullable=False),
    Column("execution_id", Integer, nullable=False),
    Column("type", Integer, nullable=False),
    Column("milliseconds_since_epoch", BigInteger),

    # Unique Constraint
    UniqueConstraint("artifact_id", "execution_id", "type", name="uniqueevent") 
)


import json
import typing as t
from cmflib.cmfquery import CmfQuery
from cmflib.cmf_merger import parse_json_to_mlmd


def identify_existing_and_new_executions(query: CmfQuery, pipeline_data: dict, pipeline_name: str) -> t.Tuple[list, list, list, str]:
    """
    Identifies and compares existing executions from a given MLMD store path with those in the MLMD payload.
    This method supports both push and pull operations by analyzing execution UUIDs and identifying overlap
    and differences between the source and target MLMD stores.

    Args:
        query (CmfQuery): The CmfQuery object.
        mlmd_data (dict): MLMD data in dictionary format.
        pipeline_name (str): The name of the pipeline.

    Returns:
        tuple: A 4-element tuple containing:
        - executions_from_path (list): All execution UUIDs found in the MLMD store at the given path.
        - list_executions_exists (list): Intersection of execution UUIDs between store and request (already present).
        - executions_from_req (list): Execution UUIDs present in the incoming MLMD payload.
        - status (str): Returns "version_update" if the payload is malformed or missing execution UUIDs.
    """
    executions_from_path = []  # Stores existing execution UUIDs from the path
    list_executions_exists: list[str] = []  # Stores the intersection of existing and new executions
    executions_from_req = []  # Extract execution UUIDs from the MLMD data payload
    status = ""

    # For metadata push → path is the server MLMD store path
    # For metadata pull → path is the client MLMD store path
    executions = query.get_all_executions_in_pipeline(pipeline_name)

    # Collect all execution UUIDs from the queried data
    for i in executions.index:
        for uuid in executions['Execution_uuid'][i].split(","):
            executions_from_path.append(uuid)

    # For metadata push → mlmd_data comes from client MLMD file
    # For metadata pull → mlmd_data comes from server MLMD file
    for stage in pipeline_data['stages']:  # checks if given execution_id present in mlmd
        for execution in stage["executions"]:
            if execution['name'] != "":  # If executions have name , they are reusable executions
                continue                # which needs to be merged in irrespective of whether already
                                        # present or not so that new artifacts associated with it gets in.
            if 'Execution_uuid' in execution['properties']:
                for uuid in execution['properties']['Execution_uuid'].split(","):
                    executions_from_req.append(uuid)
            else:
                # mlmd push is failed here
                status = "version_update"
                return executions_from_path, list_executions_exists, executions_from_req, status

    # Intersection check:
    # For metadata push → ensures only new executions get pushed
    # For metadata pull → ensures only missing executions get pull
    if executions_from_path != []:
        list_executions_exists = list(set(executions_from_path).intersection(set(executions_from_req)))
    return executions_from_path, list_executions_exists, executions_from_req, status


def update_mlmd(query: CmfQuery, req_info: str, pipeline_name: str, cmd: str, exe_uuid: str) -> str:
    """
    Updates metadata for a given pipeline by filtering out executions that already exist
    on the server and then pushing or pulling the remaining data.
    
    Args:
        query (CmfQuery): The CmfQuery object.
        req_info (str): Contains MLMD data — client-side for push, server-side for pull.  
        pipeline_name (str): The name of the pipeline to update.
        cmd (str): The command being executed, either "push" or "pull."  
        exe_uuid (str, optional): User-provided execution UUID (default: None).  
    
    Returns: A status message indicating the result of the operation:
        - "pipeline_not_exist": Pipeline does not exists inside CMF server.
        - "exists": Execution already exists on the CMF server.
        - "success": Execution successfully pushed to the CMF server.
        - "invalid_json_payload": If the JSON payload is invalid or incorrectly formatted.
        - "version_update": Mlmd push failed due to version update. 
    """
    # load the mlmd_data from the request info
    # in create executions we get full mlmd data
    mlmd_data = json.loads(req_info)
    # Ensure the pipeline name in req_info matches the one in the JSON payload to maintain data integrity
    pipelines = mlmd_data.get("Pipeline", []) # Extract "Pipeline" list, default to empty list if missing
    # pipelines contain full mlmd data with the tag - 'Pipeline'

    if not pipelines:
        return "invalid_json_payload"  # No pipelines found in payload
  
    len_pipelines = len(pipelines)
    if pipeline_name:
        # in case of push check pipeline name exists inside mlmd_data
        pipeline = [pipeline for pipeline in pipelines if pipeline.get("name") == pipeline_name]
        if not pipeline:
            return "pipeline_not_exist"

        pipeline = pipeline[0]  # Extract the first matching pipeline    

        _, list_executions_exists, _, status = identify_existing_and_new_executions(
            query, pipeline, pipeline_name
        ) 

        if status == "version_update":
            return status
        
        # remove already existing executions from the data
        for stage in pipeline['stages']:
            # Create a new filtered list to avoid modifying while iterating
            filtered_executions = []
            # Iterate through executions and remove the ones that already exist
            for cmf_exec in stage['executions']:
                uuids = cmf_exec["properties"]["Execution_uuid"].split(",")
                for uuid in uuids:
                    if not any(uuid in list_executions_exists for uuid in uuids):
                        filtered_executions.append(cmf_exec)
            stage['executions'] = filtered_executions

        # remove empty stages (those without remaining executions)
        pipeline['stages'] = [stage for stage in pipeline['stages'] if stage['executions'] != []]

        # determine if data remains to push/pull
        if len(pipeline['stages']) == 0 :
            status="exists"
        else:
            # metadata push → merge client data into server path
            # metadata pull → merge server data into client path
            if cmd == "pull":
                parse_json_to_mlmd(
                    json.dumps(pipeline), query.filepath, cmd, exe_uuid
                )
            else:
                parse_json_to_mlmd(
                    json.dumps(pipeline), "", cmd, exe_uuid
                )
            status = "success"
        return status
    else:
            for pipeline in pipelines:
                pipeline_name = pipeline.get("name")

                _, list_executions_exists, _, status = identify_existing_and_new_executions(
                    query, pipeline, pipeline_name
                )  
                if status == "version_update":
                    return status

                for stage in pipeline['stages']:
                    # Iterate through executions and remove the ones that already exist
                    for cmf_exec in stage['executions'][:]:
                        uuids = cmf_exec["properties"]["Execution_uuid"].split(",")
                        for uuid in uuids:
                            if uuid in list_executions_exists:
                                stage['executions'].remove(cmf_exec)

                # remove empty stages (those without remaining executions)
                pipeline['stages'] = [stage for stage in pipeline['stages'] if stage['executions'] != []]

                # determine if data remains to push/pull
                if len(pipeline['stages']) == 0 :
                    status="exists"
                else:
                    # metadata push → merge client data into server path
                    # metadata pull → merge server data into client path
                    if cmd == "pull":
                        parse_json_to_mlmd(
                            json.dumps(pipeline), query.filepath, cmd, exe_uuid
                        )
                    else:
                        parse_json_to_mlmd(
                            json.dumps(pipeline), "", cmd, exe_uuid
                        )
                    # we are passing this success in a very wrong way
                    status = "success"

            return status




# cmf-server api's
import io
import time
import zipfile
from fastapi import FastAPI, Request, HTTPException, Query, UploadFile, File, Depends
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import HTMLResponse, PlainTextResponse, StreamingResponse
from fastapi.staticfiles import StaticFiles
from contextlib import asynccontextmanager
import pandas as pd
from typing import List, Dict, Any, Optional
from cmflib.cmfquery import CmfQuery
import asyncio
from sqlalchemy.ext.asyncio import AsyncSession
from collections import defaultdict
from server.app.get_data import (
    get_lineage_data,
    get_mlmd_from_server,
    get_artifact_types,
    get_all_artifact_ids,
    get_all_exe_ids,
    async_api,
    get_model_data

)
from server.app.query_execution_lineage_d3force import query_execution_lineage_d3force
from server.app.query_execution_lineage_d3tree import query_execution_lineage_d3tree
from server.app.query_artifact_lineage_d3tree import query_artifact_lineage_d3tree
from server.app.query_visualization_artifact_execution import query_visualization_artifact_execution
from server.app.db.dbconfig import get_db
from server.app.db.dbqueries import (
    fetch_artifacts,
    fetch_executions,
    register_server_details,
    get_registered_server_details,
    get_sync_status,
    update_sync_status
)
from pathlib import Path
import os
import json
import typing as t
from server.app.schemas.dataframe import (
    MLMDPushRequest,
    ServerRegistrationRequest, 
    AcknowledgeRequest,
    MLMDPullRequest,
    ArtifactRequest,
    ExecutionRequest,
)
import httpx
from jsonpath_ng.ext import parse
from cmflib.cmf_federation import update_mlmd

server_store_path = "/cmf-server/data/postgres_data"
query = CmfQuery(is_server=True)

#global variables
dict_of_art_ids = {}
dict_of_exe_ids = {}
pipeline_locks = {}
lock_counts: defaultdict[str, int] = defaultdict(int)
#lifespan used to prevent multiple loading and save time for visualization.
@asynccontextmanager
async def lifespan(app: FastAPI):
    global dict_of_art_ids
    global dict_of_exe_ids
    
    if os.path.exists(server_store_path):
        # loaded execution ids with names into memory
        dict_of_exe_ids = await async_api(get_all_exe_ids, query)
        # loaded artifact ids into memory
        dict_of_art_ids = await async_api(get_all_artifact_ids, query, dict_of_exe_ids)
    yield
    dict_of_art_ids.clear()
    dict_of_exe_ids.clear()

app = FastAPI(title="cmf-server", lifespan=lifespan)

BASE_PATH = Path(__file__).resolve().parent
app.mount("/cmf-server/data/static", StaticFiles(directory="/cmf-server/data/static"), name="static")

my_ip = os.environ.get("MYIP", "127.0.0.1")
hostname = os.environ.get('HOSTNAME', "localhost")

#checking if IP or Hostname is provided,initializing url accordingly.
if my_ip != "127.0.0.1":
    url="http://"+my_ip+":3000"
else:
    url="http://"+hostname+":3000"

origins = [
    "http://localhost:3000",
    "localhost:3000",
    url
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

@app.get("/")
async def read_root(request: Request):
    return {"cmf-server"}


# api to post mlmd file to cmf-server
@app.post("/mlmd_push")
async def mlmd_push(info: MLMDPushRequest):
    print("mlmd push started")
    print("......................")
    status = "unknown_error"
    req_info = info.model_dump()  # Serializing the input data into a dictionary using model_dump()
    pipeline_name = req_info.get("pipeline_name", "")
    if pipeline_name not in pipeline_locks:    # create lock object for pipeline if it doesn't exists in lock
        pipeline_locks[pipeline_name] = asyncio.Lock()
    pipeline_lock = pipeline_locks[pipeline_name]   
    lock_counts[pipeline_name] += 1 # increment lock count by 1 if pipeline going to enter inside lock section
    async with pipeline_lock:
        try:
            status = await async_api(update_mlmd, query, req_info["json_payload"], pipeline_name, "push", req_info["exec_uuid"])
            if status == "invalid_json_payload":
                # Invalid JSON payload, return 400 Bad Request
                raise HTTPException(status_code=400, detail="Invalid JSON payload. The pipeline name is missing.")           
            if status == "version_update":
                # Raise an HTTPException with status code 422
                raise HTTPException(status_code=422, detail="version_update")
            if status != "exists":
            # async function
                await update_global_exe_dict(pipeline_name)
                await update_global_art_dict(pipeline_name)
        finally:
            lock_counts[pipeline_name] -= 1  # Decrement the reference count after lock released
            if lock_counts[pipeline_name] == 0:   #if lock_counts of pipeline is zero means lock is release from it
                del pipeline_locks[pipeline_name]  # Remove the lock if it's no longer needed
                del lock_counts[pipeline_name]
    return {"status": status}


# api to get mlmd file from cmf-server
@app.post("/mlmd_pull", response_class=HTMLResponse)
async def mlmd_pull(info: MLMDPullRequest):
    pipeline_name = info.pipeline_name
    exec_uuid = info.exec_uuid
    last_sync_time = info.last_sync_time
    print("mlmd pull started")
    print("......................")
    # checks if mlmd file exists on server
    await check_mlmd_file_exists()
    if pipeline_name:
        # checks if pipeline exists
        await check_pipeline_exists(pipeline_name)
        #json_payload values can be json data, none or no_exec_id.
        json_payload= await async_api(get_mlmd_from_server, query, pipeline_name, exec_uuid, last_sync_time, dict_of_exe_ids)
    else:
        json_payload = await async_api(get_mlmd_from_server, query, None, None, last_sync_time)

    if json_payload == None:
        raise HTTPException(status_code=406, detail=f"Pipeline {pipeline_name} not found.")
    return json_payload


@app.get("/artifacts/{pipeline_name}/{artifact_type}")
async def get_artifacts(
    pipeline_name: str, 
    artifact_type: str, 
    query_params: ArtifactRequest = Depends(),
    db: AsyncSession = Depends(get_db)
):

    filter_value = query_params.filter_value
    active_page = query_params.active_page
    sort_field = query_params.sort_field
    sort_order = query_params.sort_order
    record_per_page = query_params.record_per_page

    """Retrieve paginated artifacts with filtering, sorting, and full-text search."""
    return await fetch_artifacts(db, pipeline_name, artifact_type, filter_value, active_page, record_per_page, sort_field, sort_order)


# api to display executions available in mlmd file[from postgres]
@app.get("/executions/{pipeline_name}")
async def execution(request: Request,
                   pipeline_name: str,
                   query_params: ExecutionRequest = Depends(),
                   db: AsyncSession = Depends(get_db)
                   ):
    filter_value = query_params.filter_value
    active_page = query_params.active_page
    sort_order = query_params.sort_order
    sort_field = query_params.sort_field
    record_per_page = query_params.record_per_page

    """Retrieve paginated executions with filtering, sorting, and full-text search."""
    return await fetch_executions(db, pipeline_name, filter_value, active_page, record_per_page, sort_field, sort_order)
    

# api to display executions available in mlmd file[from postgres]
@app.get("/executions/{pipeline_name}")
async def execution(request: Request,
                   pipeline_name: str,
                   active_page: int = Query(1, description="Page number", gt=0),
                   filter_value: str = Query("", description="Search based on value"),  
                   sort_order: str = Query("asc", description="Sort by context_type(asc or desc)"),
                   db: AsyncSession = Depends(get_db)
                   ):
    """Retrieve paginated executions with filtering, sorting, and full-text search."""
    return await fetch_executions(db, pipeline_name, filter_value, active_page, 5, sort_order)
    

@app.get("/list-of-executions/{pipeline_name}")
async def list_of_executions(request: Request, pipeline_name: str):
    '''
      This api's returns list of execution types.

    '''
    # checks if mlmd file exists on server
    await check_mlmd_file_exists()
    # checks if pipeline exists
    await check_pipeline_exists(pipeline_name)
    response = await async_api(get_lineage_data, query, pipeline_name, "Execution", dict_of_art_ids, dict_of_exe_ids)
    return response

    
@app.get("/execution-lineage/tangled-tree/{uuid}/{pipeline_name}")
async def execution_lineage(request: Request, uuid: str, pipeline_name: str):
    '''
      returns dictionary of nodes and links for given execution_type.
      response = {
                   nodes: [{id:"",name:"",execution_uuid:""}],
                   links: [{source:1,target:4},{}],
                 } 
    '''
    # checks if mlmd file exists on server
    await check_mlmd_file_exists()
    # checks if pipeline exists
    await check_pipeline_exists(pipeline_name)
    response = await async_api(query_execution_lineage_d3tree, query, pipeline_name, dict_of_exe_ids, uuid)
    return response
    

@app.get("/artifact-lineage/tangled-tree/{pipeline_name}")
async def artifact_lineage_tangled(request: Request, pipeline_name: str) -> Optional[List[List[Dict[str, Any]]]]:
    '''
      Returns:
      A nested list of dictionaries with 'id' and 'parents' keys.
      response = [
        [{'id': 'data.xml.gz:236d', 'parents': []}],
        [{'id': 'parsed/train.tsv:32b7', 'parents': ['data.xml.gz:236d']}, 
        ]
    '''
    # checks if mlmd file exists on server
    await check_mlmd_file_exists()
    # checks if pipeline exists
    await check_pipeline_exists(pipeline_name)
    response = await async_api(query_artifact_lineage_d3tree, query, pipeline_name, dict_of_art_ids)        
    return response


#This api's returns list of artifact types.
@app.get("/artifact_types")
async def artifact_types():
    # checks if mlmd file exists on server
    await check_mlmd_file_exists()
    artifact_types = await async_api(get_artifact_types, query)
    if "Environment" in artifact_types:
            artifact_types.remove("Environment")
    return artifact_types


@app.get("/pipelines")
async def pipelines(request: Request):
    # checks if mlmd file exists on server
    if os.path.exists(server_store_path):
        pipeline_names = query.get_pipeline_names()
        return pipeline_names
    else:
        print("No mlmd file submitted.")
        pipeline_names = []
        return pipeline_names


@app.post("/tensorboard")
async def upload_file(request:Request, pipeline_name: str = Query(..., description="Pipeline name"),
    file: UploadFile = File(..., description="The file to upload")):
    try:
        if file.filename is None:
            raise HTTPException(status_code=400, detail="No file uploaded")
        file_path = os.path.join("/cmf-server/data/tensorboard-logs", pipeline_name, file.filename)
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "wb") as buffer:
            buffer.write(await file.read())
        return {"message": f"File '{file.filename}' uploaded successfully"}
    except Exception as e:
        return {"error": f"Failed to up load file: {e}"}


@app.get("/model-card")
async def model_card(request:Request, modelId: int, response_model=List[Dict[str, Any]]):
    json_payload_1 = ""
    json_payload_2 = ""
    json_payload_3 = ""
    json_payload_4 = ""
    model_data_df = pd.DataFrame()
    model_exe_df = pd.DataFrame()
    model_input_art_df = pd.DataFrame()
    model_output_art_df = pd.DataFrame()
    # checks if mlmd file exists on server
    await check_mlmd_file_exists()
    model_data_df, model_exe_df, model_input_art_df, model_output_art_df  = await get_model_data(query, modelId)
    if not model_data_df.empty:
        result_1 = model_data_df.to_json(orient="records")
        json_payload_1 = json.loads(result_1)
    if not model_exe_df.empty:
        result_2 = model_exe_df.to_json(orient="records")
        json_payload_2 = json.loads(result_2)
    if not model_input_art_df.empty:
        result_3 =  model_input_art_df.to_json(orient="records")
        json_payload_3 = json.loads(result_3)
    if not model_output_art_df.empty:
        result_4 =  model_output_art_df.to_json(orient="records")
        json_payload_4 = json.loads(result_4)
    return [json_payload_1, json_payload_2, json_payload_3, json_payload_4]


@app.get("/artifact-execution-lineage/tangled-tree/{pipeline_name}")
async def artifact_execution_lineage(request: Request, pipeline_name: str):
    # checks if mlmd file exists on server
    await check_mlmd_file_exists()
    # checks if pipeline exists
    await check_pipeline_exists(pipeline_name)
    response = await query_visualization_artifact_execution(query, pipeline_name, dict_of_art_ids, dict_of_exe_ids)
    return response


# Rest api is for pushing python env to upload python env
@app.post("/python-env")
async def upload_python_env(request:Request, file: UploadFile = File(..., description="The file to upload")):
    try:
        if file.filename is None:
            raise HTTPException(status_code=400, detail="No file uploaded")
        file_path = os.path.join("/cmf-server/data/env/", os.path.basename(file.filename))
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "wb") as buffer:
            buffer.write(await file.read())
        return {"message": f"File '{file.filename}' uploaded successfully"}
    except Exception as e:
        return {"error": f"Failed to up load file: {e}"}
    

# Rest api to fetch the env data from the /cmf-server/data/env folder
@app.get("/python-env", response_class=PlainTextResponse)
async def get_python_env(file_name: str) -> str:
    """
    API endpoint to fetch the content of a requirements file.

    Args:
        file_name (str): The name of the file to be fetched. Must end with .txt or .yaml.

    Returns:
        str: The content of the file as plain text.

    Raises:
        HTTPException: If the file does not exist or the extension is unsupported.
    """
    # Validate file extension
    if not (file_name.endswith(".txt") or file_name.endswith(".yaml")):
        raise HTTPException(
            status_code=400, detail="Unsupported file extension. Use .txt or .yaml"
        )
    
    # Check if the file exists
    file_path = os.path.join("/cmf-server/data/env/", os.path.basename(file_name))
    if not os.path.exists(file_path):
        raise HTTPException(status_code=404, detail="File not found")

    # Read and return the file content as plain text
    try:
        with open(file_path, "r") as file:
            content = file.read()
        return content
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Error reading file: {str(e)}")

# Rest api to push the label to /cmf-server/data/labels dir.
@app.post("/label")
async def upload_label(request:Request, file: UploadFile = File(..., description="The file to upload")):
    try:
        if not file.filename:
            raise HTTPException(status_code=400, detail="No file provided.")
        
        # Construct the full file path in /server/data/labels
        labels_dir = "/cmf-server/data/labels"
        file_path = os.path.join(labels_dir, os.path.basename(file.filename))
        
        # Ensure the directory exists
        os.makedirs(os.path.dirname(file_path), exist_ok=True)

        # Check if the file already exists
        if os.path.exists(file_path):
            return {"message": f"File '{file.filename}' already exists at {labels_dir}. Skipping upload."}


        # Save the uploaded file
        with open(file_path, "wb") as buffer:
            buffer.write(await file.read())

        return {"message": f"File '{file.filename}' uploaded successfully to {labels_dir}."}
    
    except Exception as e:
        return {"error": f"Failed to up load file: {e}"}
    

# Rest api to fetch the label data from the /cmf-server/data/labels folder
@app.get("/label-data", response_class=PlainTextResponse)
async def get_label_data(file_name: str) -> str:
    """
    API endpoint to fetch the content of a requirements file.

    Args:
        file_name (str): The name of the file to be fetched. Must end with .csv.

    Returns:
        str: The content of the file as plain text.

    Raises:
        HTTPException: If the file does not exist or the extension is unsupported.
    """
    
    # Check if the file exists
    file_path = os.path.join("/cmf-server/data/labels/", os.path.basename(file_name))
    if not os.path.exists(file_path):
        raise HTTPException(status_code=404, detail="File not found")

    # Read and return the file content as plain text
    try:
        with open(file_path, "r") as file:
            content = file.read()
        return content
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Error reading file: {str(e)}")


@app.post("/register-server")
async def register_server(request: ServerRegistrationRequest, db: AsyncSession = Depends(get_db)):
    try:
        # Access the data from the Pydantic model
        server_name = request.server_name
        host_info = request.host_info

        # Step 1: Send a request to the target server
        async with httpx.AsyncClient() as client:
            try:
                response = await client.post(
                    f"http://{host_info}:8080/acknowledge",
                    json={"server_name": server_name, "host_info": host_info}
                )
                if response.status_code != 200:
                    raise HTTPException(status_code=500, detail="Target server did not respond successfully")
                target_server_data = response.json()
            except httpx.RequestError:
                raise HTTPException(status_code=500, detail="Target server is not reachable")

        # Check user is registring with own details
        if host_info in [my_ip, hostname, "127.0.0.1", "localhost"]:
            # Restrict the user from registering with own details
            return {"message": "Registration failed: Cannot register the server with its own details."}

        return await register_server_details(db, server_name, host_info)
    
    except HTTPException as e:
        # Re-raise known error without wrapping
        raise e

    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Failed to register server: {e}")
    

@app.post("/acknowledge")
async def acknowledge(request: AcknowledgeRequest):
    # Acknowledge the connection setup
    return {
        "message": f"Hi, I acknowledge your request.",
    }


async def server_mlmd_pull(host_info, last_sync_time):
    """
    Fetch mlmd data from a specified server.

    Args:
        host_info (str): The host information (IP or hostname) of the target server.
        last_sync_time (int): The last sync time in milliseconds since epoch.

    Returns:
        dict: The mlmd data fetched from the specified server.

    Raises:
        HTTPException: If the server is not reachable or an error occurs during the request.
    """
    try:
        # Step 1: Send a request to the target server to fetch mlmd data
        async with httpx.AsyncClient(timeout=300.0) as client:
            try:
                response = await client.post(f"http://{host_info}:8080/mlmd_pull", json={'last_sync_time': last_sync_time})

                if response.status_code != 200:
                    raise HTTPException(status_code=500, detail="Target server did not respond successfully")

                json_payload = response.json()
                python_env_store_path = "/cmf-server/data/env"

                if last_sync_time:
                    # Extract the Environment file names from the JSON payload
                    data = json_payload
                    jsonpath_expr = parse('$..events[?(@.artifact.type == "Environment")].artifact.name')
                    environment_names = {match.value.split(":")[0].split("/")[-1] for match in jsonpath_expr.find(data)}

                    # Check if the list is empty or not
                    # if list is empty then no need to download the zip file
                    if len(environment_names) == 0: 
                        print("No Environment files are found inside json payload.")
                        return json_payload

                    list_of_files = list(environment_names)
                    # Added list_of_files to the request as a optional query parameter 
                    python_env_zip = await client.get(f"http://{host_info}:8080/download-python-env", params=list_of_files)
                else:
                    python_env_zip = await client.get(f"http://{host_info}:8080/download-python-env", params=None)

                if python_env_zip.status_code == 200:
                    try:
                        # Create the directory if it doesn't exist
                        os.makedirs(python_env_store_path, exist_ok=True)

                        # Unzip the zip file content
                        with zipfile.ZipFile(io.BytesIO(python_env_zip.content)) as zf:
                            # Extract all files to a temporary directory
                            temp_dir = os.path.join(python_env_store_path, "temp_extracted")
                            os.makedirs(temp_dir, exist_ok=True)
                            zf.extractall(temp_dir)

                            # Move all extracted files to the target directory
                            for root, dirs, files in os.walk(temp_dir):
                                for file in files:
                                    src_file = os.path.join(root, file)
                                    dest_file = os.path.join(python_env_store_path, file)
                                    try:
                                        os.rename(src_file, dest_file)
                                        print(f"Moved {src_file} to {dest_file}")
                                    except Exception as e:
                                        print(f"Failed to move {src_file} to {dest_file}: {e}")

                            # Clean up the temporary directory
                            os.rmdir(temp_dir)
                        # print("Storing at:", os.path.abspath(python_env_store_path))
                        # print("Files in target directory:", os.listdir(python_env_store_path))
                        print("All files stored successfully.")
                    except Exception as e:
                        print(f"Error during file extraction or storage: {e}")
                else:
                    print(f"Failed to download ZIP file. Status code: {python_env_zip.status_code}")
            except httpx.RequestError:
                raise HTTPException(status_code=500, detail="Target server is not reachable")
        return json_payload

    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Failed to fetch mlmd data: {e}")


@app.post("/sync")
async def sync_metadata(request: ServerRegistrationRequest, db: AsyncSession = Depends(get_db)):
    """
    Synchronize metadata for a registered server.

    Args:
        request (ServerRegistrationRequest): The request containing server details.

    Returns:
        dict: A response containing the sync status and last sync time.

    Raises:
        HTTPException: If the server is not found or an error occurs during synchronization.
    """
    try:
        # Access the data from the Pydantic model
        server_name = request.server_name
        host_info = request.host_info

        # Fetch the server details from the database
        row = await get_sync_status(db, server_name, host_info)

        if not row:
            raise HTTPException(status_code=404, detail="Server not found in the registered servers list")

        last_sync_time = row[0]['last_sync_time']
        current_utc_epoch_time = int(time.time() * 1000)

        # Call the function to fetch the JSON payload
        json_payload = await server_mlmd_pull(host_info, last_sync_time)      


        # Use the JSON payload in json_data
        json_data = {
            "exec_uuid": None,
            "json_payload": json.dumps(json_payload),
            "pipeline_name": None
        }

        # Ensure the pipeline name in req_info matches the one in the JSON payload to maintain data integrity
        pipelines = json_payload.get("Pipeline", []) # Extract "Pipeline" list, default to empty list if missing
        # pipelines contain full mlmd data with the tag - 'Pipeline'
        # print("type of pipelines = ", type(pipelines))
        # tell us number of pipelines
        len_pipelines = len(pipelines)
        pipeline_names = []

        if not pipelines:
            return {
                "message": "Nothing to sync",
                "status": "success",
                "last_sync_time": current_utc_epoch_time
            }

        # in case of push check pipeline name exists inside mlmd_data
        pipeline_names = [pipeline.get("name") for pipeline in pipelines]

        # Push the JSON payload to the host server
        status = await async_api(update_mlmd, query, json_data["json_payload"], None, "push", None)
        if status == "invalid_json_payload":
            # Invalid JSON payload, return 400 Bad Request
            raise HTTPException(status_code=400, detail="Invalid JSON payload. The pipeline name is missing.")           
        if status == "version_update":
            # Raise an HTTPException with status code 422
            raise HTTPException(status_code=422, detail="version_update")
        global dict_of_art_ids
        global dict_of_exe_ids
        message = "Nothing to sync."
        if status != "exists":
            if not last_sync_time:
                # this is not completely correct 
                # as before we do the sync first time, it is entirely possible that there exists a 
                # pipeline on the server 1  - test this scenario
                message = f"Host server is syncing with the selected server '{server_name}' at address '{host_info}' for the first time."
                for pipeline_name in pipeline_names:
                    dict_of_exe_ids = get_all_exe_ids(query)
                    dict_of_art_ids = get_all_artifact_ids(query, dict_of_exe_ids)
            else:
                message = f"Host server is being synced with the selected server '{server_name}' at address '{host_info}'."
                for pipeline_name in pipeline_names:
                    update_global_exe_dict(pipeline_name)
                    update_global_art_dict(pipeline_name)

        # Update the last_sync_time in the database only if sync status is successful
        if status == "success":
            await update_sync_status(db, current_utc_epoch_time, server_name, host_info)
        return {
            "message": message,
            "status": status,
            "last_sync_time": current_utc_epoch_time
        }

    except Exception as e:
        print(e)
        raise HTTPException(status_code=500, detail=f"Failed to sync metadata: {e}")


@app.get("/server-list")
async def server_list(db: AsyncSession = Depends(get_db)):
    rows = await get_registered_server_details(db)
    return rows


@app.get("/download-python-env")
def download_python_env(request: Request, list_of_files: Optional[list[str]] = Query(None)):
    """
    API endpoint to compress and download the entire folder as a ZIP file.
    """
    try:
        DIRECTORY = "/cmf-server/data/env/"  # Directory to be compressed
        # Check if the directory exists
        if not os.path.exists(DIRECTORY):
            return {"error": "Directory does not exist"}

        # Determine files to include in the ZIP
        files_to_zip = []
        # if list_of_files is provided, include only those files
        # else include all files in the directory
        if list_of_files:
            for file_name in list_of_files:
                file_path = os.path.join(DIRECTORY, file_name)
                if os.path.exists(file_path):
                    files_to_zip.append((file_path, file_name))
                else:
                    return {"error": f"File {file_name} does not exist"}
        else:
            if not os.listdir(DIRECTORY):
                return {"error": "Directory is empty"}
            for root, _, files in os.walk(DIRECTORY):
                for file in files:
                    file_path = os.path.join(root, file)
                    arcname = os.path.relpath(file_path, DIRECTORY)
                    files_to_zip.append((file_path, arcname))

        # Create and send the ZIP file 
        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
            for file_path, arcname in files_to_zip:
                zip_file.write(file_path, arcname)
        
        zip_buffer.seek(0)

        return StreamingResponse(
            zip_buffer,
            media_type="application/zip",
            headers={
                "Content-Disposition": f"attachment; filename={'python_env_files.zip' if list_of_files else 'python_env_folder.zip'}"
            }
        )
    except Exception as e:
        return {"error": str(e)}


async def update_global_art_dict(pipeline_name):
    global dict_of_art_ids
    output_dict = await async_api(get_all_artifact_ids, query, dict_of_exe_ids, pipeline_name)
    dict_of_art_ids[pipeline_name]=output_dict[pipeline_name]
    return


async def update_global_exe_dict(pipeline_name):
    global dict_of_exe_ids
    output_dict = await async_api(get_all_exe_ids, query, pipeline_name)
    dict_of_exe_ids[pipeline_name] = output_dict[pipeline_name]  
    return


# Function to checks if mlmd file exists on server
async def check_mlmd_file_exists():
    if not os.path.exists(server_store_path):
        print(f"{server_store_path} file doesn't exist.")
        raise HTTPException(status_code=404, detail=f"{server_store_path} file doesn't exist.")


# Function to check if the pipeline exists
async def check_pipeline_exists(pipeline_name):
    if pipeline_name not in query.get_pipeline_names():
        print(f"Pipeline {pipeline_name} not found.")
        raise HTTPException(status_code=404, detail=f"Pipeline {pipeline_name} not found.")


"""
This API is no longer in use within the project but is retained for reference or potential future use.
@app.get("/execution-lineage/force-directed-graph/{pipeline_name}/{uuid}")
async def execution_lineage(request: Request, pipeline_name: str, uuid: str):
    '''
      returns dictionary of nodes and links for given execution_type.
      response = {
                   nodes: [{id:"",name:"",execution_uuid:""}],
                   links: [{source:1,target:4},{}],
                 } 
    '''
    # checks if mlmd file exists on server
    if os.path.exists(server_store_path):
        query = cmfquery.CmfQuery(server_store_path)
        if (pipeline_name in query.get_pipeline_names()):
            response = await async_api(query_execution_lineage_d3force, server_store_path, pipeline_name, dict_of_exe_ids, uuid)
    else:
        response = None
    return response


@app.get("/artifact-lineage/force-directed-graph/{pipeline_name}")
async def artifact_lineage(request: Request, pipeline_name: str):
    '''
      This api returns dictionary of nodes and links for given pipeline.
      response = {
                   nodes: [{id:"",name:""}],
                   links: [{source:1,target:4},{}],
                 }

    '''
    # checks if mlmd file exists on server
    if os.path.exists(server_store_path):
        query = cmfquery.CmfQuery(server_store_path)
        if (pipeline_name in query.get_pipeline_names()):
            response=await async_api(get_lineage_data, server_store_path, pipeline_name, "Artifacts", dict_of_art_ids, dict_of_exe_ids)
            return response
        else:
            return f"Pipeline name {pipeline_name} doesn't exist."

    else:
        return None
"""
